import sys

import pytest
from packaging.version import Version

from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseConfigError

from .test_GlobalConfig import VALID_VERSIONS


def test_minimal_AppConfig():
    """A simple config can be defined."""
    config = AppConfig(
        app_name="myapp",
        version="1.2.3",
        bundle="org.beeware",
        description="A simple app",
        sources=["src/myapp", "somewhere/else/interesting", "local_app"],
        license={"file": "LICENSE"},
    )

    # The basic properties have been set.
    assert config.app_name == "myapp"
    assert config.version == Version("1.2.3")
    assert config.bundle == "org.beeware"
    assert config.description == "A simple app"
    assert config.requires is None
    assert config.sources == ["src/myapp", "somewhere/else/interesting", "local_app"]
    assert config.external_package_path is None
    assert config.external_package_executable_path is None

    # Derived properties have been set.
    assert config.bundle_name == "myapp"
    assert config.bundle_identifier == "org.beeware.myapp"
    assert config.formal_name == "myapp"
    assert config.class_name == "myapp"
    assert config.document_types == {}

    # There is no icon of any kind
    assert config.icon is None

    # The PYTHONPATH is derived correctly
    config.test_mode = False
    assert config.PYTHONPATH() == ["src", "somewhere/else", ""]
    # The test mode PYTHONPATH is the same
    config.test_mode = True
    assert config.PYTHONPATH() == ["src", "somewhere/else", ""]

    # The object has a meaningful REPL
    assert repr(config) == "<org.beeware.myapp v1.2.3 AppConfig>"


def test_minimal_external_AppConfig():
    """A simple config for an external app can be defined."""
    config = AppConfig(
        app_name="myapp",
        version="1.2.3",
        bundle="org.beeware",
        description="A simple app",
        license={"file": "LICENSE"},
        external_package_path="path/to/package",
        external_package_executable_path="internal/app.exe",
    )

    # The basic properties have been set.
    assert config.app_name == "myapp"
    assert config.version == Version("1.2.3")
    assert config.bundle == "org.beeware"
    assert config.description == "A simple app"
    assert config.requires is None
    assert config.sources is None
    assert config.external_package_path == "path/to/package"
    assert config.external_package_executable_path == "internal/app.exe"

    # Derived properties have been set.
    assert config.bundle_name == "myapp"
    assert config.bundle_identifier == "org.beeware.myapp"
    assert config.formal_name == "myapp"
    assert config.class_name == "myapp"
    assert config.document_types == {}

    # There is no icon of any kind
    assert config.icon is None

    # The PYTHONPATH is derived correctly
    config.test_mode = False
    assert config.PYTHONPATH() == []
    # The test mode PYTHONPATH is the same
    config.test_mode = True
    assert config.PYTHONPATH() == []

    # The object has a meaningful REPL
    assert repr(config) == "<org.beeware.myapp v1.2.3 AppConfig>"


def test_extra_attrs():
    """A config can contain attributes in addition to those required."""
    config = AppConfig(
        app_name="myapp",
        formal_name="My App!",
        version="1.2.3",
        bundle="org.beeware",
        description="A simple app",
        long_description="A longer description\nof the app",
        license={"file": "LICENSE"},
        template="/path/to/template",
        sources=["src/myapp"],
        requires=["first", "second", "third"],
        document_type={
            "document": {
                "icon": "icon",
                "extension": "doc",
                "description": "A document",
                "url": "https://testurl.com",
                "mime_type": "application/x-my-doc-type",
            }
        },
        first="value 1",
        second=42,
    )

    # The basic properties have been set.
    assert config.app_name == "myapp"
    assert config.version == Version("1.2.3")
    assert config.bundle == "org.beeware"
    assert config.description == "A simple app"
    assert config.long_description == "A longer description\nof the app"
    assert config.template == "/path/to/template"
    assert config.requires == ["first", "second", "third"]

    # Properties that are derived by default have been set explicitly
    assert config.formal_name == "My App!"
    assert config.class_name == "MyApp"

    if sys.platform == "darwin":
        assert config.document_types == {
            "document": {
                "icon": "icon",
                "extension": "doc",
                "description": "A document",
                "url": "https://testurl.com",
                "mime_type": "application/x-my-doc-type",
                "macOS": {
                    "CFBundleTypeRole": "Viewer",
                    "LSHandlerRank": "Owner",
                    "UTTypeConformsTo": [
                        "public.data",
                        "public.content",
                    ],
                    "is_core_type": False,
                },
            }
        }
    else:
        assert config.document_types == {
            "document": {
                "icon": "icon",
                "extension": "doc",
                "description": "A document",
                "url": "https://testurl.com",
                "mime_type": "application/x-my-doc-type",
            }
        }

    # Explicit additional properties have been set
    assert config.first == "value 1"
    assert config.second == 42

    # An attribute that wasn't provided raises an error
    with pytest.raises(AttributeError):
        _ = config.unknown


@pytest.mark.parametrize(
    "name",
    [
        "myapp",  # lowercase
        "myApp",  # contains uppercase
        "MyApp",  # initial uppercase
        "MyAPP",  # ends in uppercase
        "my-app",  # contains hyphen
        "my_app",  # contains underscore
        "myapp2",  # ends with digit
        "my2app",  # contains digit
    ],
)
def test_valid_app_name(name):
    try:
        AppConfig(
            app_name=name,
            version="1.2.3",
            bundle="org.beeware",
            description="A simple app",
            sources=["src/" + name.replace("-", "_")],
            license={"file": "LICENSE"},
        )
    except BriefcaseConfigError:
        pytest.fail(f"{name} should be valid")


@pytest.mark.parametrize(
    "name",
    [
        "!myapp",  # initial punctuation
        "my!app",  # contains punctuation
        "myapp!",  # end punctuation
        "my$app",  # other punctuation
        "-myApp",  # initial hyphen
        "myApp-",  # end hyphen
        "_myApp",  # initial underscore
        "myApp_",  # end underscore
    ],
)
def test_invalid_app_name(name):
    with pytest.raises(BriefcaseConfigError, match=r"is not a valid app name\."):
        AppConfig(
            app_name=name,
            version="1.2.3",
            bundle="org.beeware",
            description="A simple app",
            sources=["src/invalid"],
            license={"file": "LICENSE"},
        )


@pytest.mark.parametrize(
    "bundle",
    [
        "com.example",
        "com.example.more",
        "com.example42.more",
        "com.example-42.more",
    ],
)
def test_valid_bundle(bundle):
    try:
        AppConfig(
            app_name="myapp",
            version="1.2.3",
            bundle=bundle,
            description="A simple app",
            sources=["src/myapp"],
            license={"file": "LICENSE"},
        )
    except BriefcaseConfigError:
        pytest.fail(f"{bundle} should be valid")


@pytest.mark.parametrize(
    "bundle",
    [
        "not a bundle!",  # Free text.
        "home",  # Only one section.
        "com.hello_world",  # underscore
        "com.hello,world",  # comma
        "com.hello world!",  # exclamation point
    ],
)
def test_invalid_bundle_identifier(bundle):
    with pytest.raises(
        BriefcaseConfigError, match=r"is not a valid bundle identifier\."
    ):
        AppConfig(
            app_name="myapp",
            version="1.2.3",
            bundle=bundle,
            description="A simple app",
            sources=["src/invalid"],
            license={"file": "LICENSE"},
        )


@pytest.mark.parametrize(("input", "expected"), VALID_VERSIONS)
def test_valid_app_version(input, expected):
    config = AppConfig(
        app_name="myapp",
        version=input,
        bundle="org.beeware",
        description="A simple app",
        sources=["src/myapp"],
        license={"file": "LICENSE"},
    )

    # Version is parsed as an equivalent Version object
    assert config.version == Version(expected)
    # Serialized output of parsed Version matches the expected value
    assert str(config.version) == expected


def test_invalid_app_version():
    with pytest.raises(
        BriefcaseConfigError,
        match=r"Version number for 'myapp' \(foobar\) is not valid\.",
    ):
        AppConfig(
            app_name="myapp",
            version="foobar",
            bundle="org.beeware",
            description="A simple app",
            sources=["src/invalid"],
            license={"file": "LICENSE"},
        )


@pytest.mark.parametrize(
    ("name", "module_name"),
    [
        ("myapp", "myapp"),
        ("my-app", "my_app"),
    ],
)
def test_module_name(name, module_name):
    config = AppConfig(
        app_name=name,
        version="1.2.3",
        bundle="org.beeware",
        description="A simple app",
        sources=["src/" + module_name],
        license={"file": "LICENSE"},
    )

    assert config.module_name == module_name


@pytest.mark.parametrize(
    ("bundle", "package_name"),
    [
        ("com.example", "com.example"),
        ("com.ex-ample", "com.ex_ample"),
    ],
)
def test_package_name(bundle, package_name):
    config = AppConfig(
        app_name="myapp",
        version="1.2.3",
        bundle=bundle,
        description="A simple app",
        sources=["src/myapp"],
        license={"file": "LICENSE"},
    )

    assert config.package_name == package_name


@pytest.mark.parametrize(
    ("app_name", "dist_info_name"),
    [
        ("my-app", "my_app.dist-info"),
        ("my_app", "my_app.dist-info"),
    ],
)
def test_dist_info_name(app_name, dist_info_name):
    config = AppConfig(
        app_name=app_name,
        version="1.2.3",
        bundle="com.example",
        description="A simple app",
        sources=["src/my_app"],
        license={"file": "LICENSE"},
    )

    assert config.dist_info_name == dist_info_name


@pytest.mark.parametrize(
    ("app_name", "bundle_name"),
    [
        ("my-app", "my-app"),
        ("my_app", "my-app"),
    ],
)
def test_bundle_name(app_name, bundle_name):
    config = AppConfig(
        app_name=app_name,
        version="1.2.3",
        bundle="com.example",
        description="A simple app",
        sources=["src/my_app"],
        license={"file": "LICENSE"},
    )

    assert config.bundle_name == bundle_name


@pytest.mark.parametrize(
    ("app_name", "bundle_name"),
    [
        ("my-app", "my-app"),
        ("my_app", "my-app"),
    ],
)
def test_bundle_identifier(app_name, bundle_name):
    bundle = "com.example"

    config = AppConfig(
        app_name=app_name,
        version="1.2.3",
        bundle=bundle,
        description="A simple app",
        sources=["src/my_app"],
        license={"file": "LICENSE"},
    )

    assert config.bundle_identifier == f"{bundle}.{bundle_name}"


@pytest.mark.parametrize(
    "sources",
    [
        ["src/dupe", "src/dupe"],
        ["src/dupe", "src/other", "src/dupe"],
        ["src/dupe", "somewhere/dupe", "src/other"],
        ["src/dupe", "src/deep/dupe", "src/other"],
    ],
)
def test_duplicated_source(sources):
    with pytest.raises(
        BriefcaseConfigError, match=r"contains duplicated package names\."
    ):
        AppConfig(
            app_name="dupe",
            version="1.2.3",
            bundle="org.beeware",
            description="A simple app",
            sources=sources,
            license={"file": "LICENSE"},
        )


def test_no_source_for_app():
    with pytest.raises(
        BriefcaseConfigError, match=r" does not include a package named 'my_app'\."
    ):
        AppConfig(
            app_name="my-app",
            version="1.2.3",
            bundle="org.beeware",
            description="A simple app",
            sources=["src/something", "src/other"],
            license={"file": "LICENSE"},
        )


@pytest.mark.parametrize(
    ("install_launcher", "console_app", "expected"),
    [
        (True, False, True),
        (False, False, False),
        (None, False, True),
        #
        (True, True, True),
        (False, True, False),
        (None, True, False),
    ],
)
def test_install_launcher(install_launcher, console_app, expected):
    config = AppConfig(
        app_name="my-app",
        version="1.2.3",
        bundle="org.beeware",
        description="A simple app",
        license={"file": "LICENSE"},
        install_launcher=install_launcher,
        console_app=console_app,
    )
    assert config.install_launcher == expected


def test_non_unique_uninstall_options():
    """Installer and Uninstaller options must not overlap."""
    with pytest.raises(
        BriefcaseConfigError,
        match=(
            r"Uininstall option names must be unique. "
            r"The name 'first' is already used as an install option."
        ),
    ):
        AppConfig(
            app_name="myapp",
            version="1.2.3",
            bundle="org.beeware",
            description="A simple app",
            sources=["src/myapp"],
            license={"file": "LICENSE"},
            install_option=[
                {
                    "name": "first",
                    "title": "First option",
                    "description": "Do the first thing",
                    "default": True,
                },
                {
                    "name": "second",
                    "title": "Second option",
                    "description": "Do the second thing",
                    "default": False,
                },
            ],
            uninstall_option=[
                {
                    "name": "first",
                    "title": "First option",
                    "description": "Do the first thing",
                    "default": True,
                },
            ],
        )


def test_capitalization():
    """Capitalization is prohibited and normalized out in some properties."""
    config = AppConfig(
        app_name="MyApp",
        version="1.2.3",
        bundle="Org.Beeware",
        description="A simple app",
        sources=["src/MyApp", "somewhere/else/interesting", "local_app"],
        license={"file": "LICENSE"},
    )

    # The basic properties have been set.
    assert config.app_name == "MyApp"
    assert config.bundle == "org.beeware"

    # Derived properties have been set.
    assert config.bundle_name == "myapp"
    assert config.bundle_identifier == "org.beeware.myapp"
    assert config.formal_name == "MyApp"
    assert config.class_name == "MyApp"

    # The object has a meaningful REPL
    assert repr(config) == "<org.beeware.myapp v1.2.3 AppConfig>"



import pytest

from briefcase.commands.create import Git
from briefcase.exceptions import BriefcaseCommandError


def test_no_git(update_command, monkeypatch):
    """If Git is not installed, an error is raised."""

    def monkeypatch_verify_git(*a, **kw):
        raise BriefcaseCommandError("Briefcase requires git, but it is not installed")

    monkeypatch.setattr(Git, "verify", monkeypatch_verify_git)

    # The command will fail tool verification.
    with pytest.raises(
        BriefcaseCommandError, match=r"Briefcase requires git, but it is not installed"
    ):
        update_command()


def test_update(update_command, first_app, second_app):
    """The update command can be called."""
    # Configure no command line options
    options, _ = update_command.parse_options([])

    update_command(**options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup", "first"),
        # Update the second app
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("code", "second", False),
        ("cleanup", "second"),
    ]


def test_update_single(update_command, first_app, second_app):
    """The update command can be called to update a single app from the config."""
    # Configure no command line options
    options, _ = update_command.parse_options([])

    update_command(app=update_command.apps["first"], **options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # Update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup", "first"),
    ]


def test_update_with_requirements(update_command, first_app, second_app):
    """The update command can be called, requesting a requirements update."""
    # Configure a requirements update
    options, _ = update_command.parse_options(["-r"])

    update_command(**options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("requirements", "first", False, False),
        ("cleanup", "first"),
        # Update the second app
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("code", "second", False),
        ("requirements", "second", False, False),
        ("cleanup", "second"),
    ]


def test_update_with_resources(update_command, first_app, second_app):
    """The update command can be called, requesting a resources update."""
    # Configure no command line options
    options, _ = update_command.parse_options(["--update-resources"])

    update_command(**options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("resources", "first"),
        ("cleanup", "first"),
        # Update the second app
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("code", "second", False),
        ("resources", "second"),
        ("cleanup", "second"),
    ]


def test_update_with_support(update_command, first_app, second_app):
    """The update command can be called, requesting an app support update."""
    # Configure no command line options
    options, _ = update_command.parse_options(["--update-support"])

    update_command(**options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup-support", "first"),
        ("support", "first"),
        ("cleanup", "first"),
        # Update the second app
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("code", "second", False),
        ("cleanup-support", "second"),
        ("support", "second"),
        ("cleanup", "second"),
    ]


# Parametrize both --apps/-a flags
@pytest.mark.parametrize("app_flags", ["--app", "-a"])
def test_update_app_single(update_command, first_app, second_app, app_flags):
    """If the --app or -a flag is used, only the selected app is updated."""
    # Configure command line options with the parameterized flag
    options, _ = update_command.parse_options([app_flags, "first"])

    # Run the update command
    update_command(**options)

    # Only the selected app is updated
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # Update the app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup", "first"),
    ]


def test_update_app_invalid(update_command, first_app, second_app):
    """If an invalid app name is passed to --app, raise an error."""
    # Add two apps
    update_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure the --app option with an invalid app
    options, _ = update_command.parse_options(["--app", "invalid"])

    # Running the command should raise an error
    with pytest.raises(
        BriefcaseCommandError,
        match=r"App 'invalid' does not exist in this project.",
    ):
        update_command(**options)


def test_update_app_none_defined(update_command):
    """If no apps are defined, do nothing."""
    # No apps defined
    update_command.apps = {}

    # Configure no command line options
    options, _ = update_command.parse_options([])

    # Run the update command
    result = update_command(**options)

    # The right sequence of things will be done
    assert result is None
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
    ]


def test_update_app_all_flags(update_command, first_app, second_app):
    """Verify that all update-related flags work correctly with -a."""
    options, _ = update_command.parse_options(
        [
            "-a",
            "first",
            "--update-requirements",
            "--update-resources",
            "--update-support",
            "--update-stub",
            "--no-input",
            "--log",
            "-vv",
        ]
    )

    # Run the update command
    update_command(**options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # Update the app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("requirements", "first", False, False),
        ("resources", "first"),
        ("cleanup-support", "first"),
        ("support", "first"),
        ("cleanup", "first"),
    ]


def test_update_external_app(update_command, first_app):
    """If the user requests a update an external app, an error is raised."""
    # Add an apps
    update_command.apps = {
        "first": first_app,
    }

    # Make first_app an external app
    first_app.sources = None
    first_app.external_package_path = "path/to/package"

    # Configure no command line options
    options, _ = update_command.parse_options([])

    # Run the build command
    with pytest.raises(
        BriefcaseCommandError,
        match=r"'first' is declared as an external app",
    ):
        update_command(**options)


def test_update_debug_with_requirements(update_command, first_app, second_app):
    """The update command can be called, requesting a requirements update."""
    update_command.supports_debugger = True

    # Configure no command line options
    options, _ = update_command.parse_options(["-r", "--debug=pdb"])

    update_command(**options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("requirements", "first", False, True),
        ("cleanup", "first"),
        # Update the second app
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("code", "second", False),
        ("requirements", "second", False, True),
        ("cleanup", "second"),
    ]



import importlib.machinery


def test_venv_name(monkeypatch, dev_command):
    """Name of the venv is platform specific."""
    # Set up some fake platform details
    monkeypatch.setattr(
        importlib.machinery,
        "EXTENSION_SUFFIXES",
        [".cpython-3X-gothic.so", ".cpython-abi3.so", ".so"],
    )

    assert dev_command.venv_name == "dev.cpython-3X-gothic"



import pytest

from briefcase.integrations.file import File

from ...utils import create_file


@pytest.mark.parametrize(
    ("files", "sorted"),
    [
        # Files in a directory are sorted lexically
        (
            ["foo/bar/a.txt", "foo/bar/c.txt", "foo/bar/b.txt"],
            ["foo/bar/c.txt", "foo/bar/b.txt", "foo/bar/a.txt"],
        ),
        # Subfolders are sorted before files in that directory; but sorted lexically in themselves
        (
            [
                "foo/bar/b",
                "foo/bar/b/aaa.txt",
                "foo/bar/b/zzz.txt",
                "foo/bar/b/deeper",
                "foo/bar/b/deeper/deeper_db2.txt",
                "foo/bar/b/deeper/deeper_db1.txt",
                "foo/bar/a.txt",
                "foo/bar/c.txt",
                "foo/bar/e.txt",
                "foo/bar/d",
                "foo/bar/d/deep_d2.txt",
                "foo/bar/d/deep_d1.txt",
            ],
            [
                "foo/bar/d/deep_d2.txt",
                "foo/bar/d/deep_d1.txt",
                "foo/bar/b/deeper/deeper_db2.txt",
                "foo/bar/b/deeper/deeper_db1.txt",
                "foo/bar/b/deeper",
                "foo/bar/b/zzz.txt",
                "foo/bar/b/aaa.txt",
                "foo/bar/d",
                "foo/bar/b",
                "foo/bar/e.txt",
                "foo/bar/c.txt",
                "foo/bar/a.txt",
            ],
        ),
        # If a folder contains both folders and files, the folders are returned first.
        (
            [
                "foo/bar/a",
                "foo/bar/b.txt",
                "foo/bar/c",
            ],
            [
                "foo/bar/c",
                "foo/bar/a",
                "foo/bar/b.txt",
            ],
        ),
    ],
)
def test_sorted_depth_first(files, sorted, tmp_path):
    # Convert the strings into paths in the temp folder
    paths = [tmp_path / file_path for file_path in files]

    # Create all the paths that have a suffix
    for file_path in paths:
        if file_path.suffix:
            create_file(tmp_path / file_path, content=str(file_path))
        else:
            file_path.mkdir(parents=True, exist_ok=True)

    assert File.sorted_depth_first(paths) == [
        tmp_path / file_path for file_path in sorted
    ]



from unittest import mock

import pytest

from briefcase.commands import DevCommand
from briefcase.commands.base import full_options
from briefcase.exceptions import BriefcaseCommandError, RequirementsInstallError
from briefcase.integrations.virtual_environment import VenvContext


class DummyDevCommand(DevCommand):
    """A dummy Dev command that doesn't actually do anything.

    It only serves to track which actions would be performed.
    """

    # Platform and format contain upper case to test case normalization
    platform = "Tester"
    output_format = "Dummy"
    description = "Dummy dev command"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, apps={}, **kwargs)

        self.actions = []
        self.env = {"a": 1, "b": 2, "c": 3}

        self.tools.virtual_environment.create = mock.MagicMock(
            side_effect=self.virtual_environment
        )
        # Track which venvs exist for this command instance
        self.venvs = {}

    def simulate_existing_venv(self, appname, isolated=True):
        """Mark a venv as already existing (for testing installed apps)."""
        mock_venv_context = mock.MagicMock(spec=VenvContext)
        mock_venv_context.created = False
        self.venvs[(appname, isolated)] = mock_venv_context
        return mock_venv_context

    def verify_host(self):
        super().verify_host()
        self.actions.append(("verify-host",))

    def verify_tools(self):
        super().verify_tools()
        self.actions.append(("verify-tools",))

    def verify_app_template(self, app):
        super().verify_app_template(app=app)
        self.actions.append(("verify-app-template", app.app_name))

    def verify_app_tools(self, app):
        super().verify_app_tools(app=app)
        self.actions.append(("verify-app-tools", app.app_name))

    def virtual_environment(self, venv_path, isolated=False, recreate=False):
        # Extract the app name from the venv_path to use as a cache key.
        app_name = venv_path.parts[-2]
        self.actions.append(("virtual-environment", app_name, isolated, recreate))

        # Simulate venv.created behavior:
        # - created=True if venv doesn't exist yet (first time) OR recreate=True
        # - created=False if venv already exists and recreate=False

        venv_key = (app_name, isolated)

        try:
            mock_venv_context = self.venvs[venv_key]
            # Venv already exists; if we're recreating, flag that
            mock_venv_context.created = recreate
        except KeyError:
            # First time creating this venv
            mock_venv_context = mock.MagicMock(spec=VenvContext)
            mock_venv_context.created = True
            self.venvs[venv_key] = mock_venv_context

        mock_context = mock.MagicMock()
        mock_context.__enter__.return_value = mock_venv_context
        mock_context.__exit__.return_value = False
        return mock_context

    def install_dev_requirements(self, app, venv, **kwargs):
        self.actions.append(("dev_requirements", app.app_name, venv))

    def get_environment(self, app):
        return self.env

    def run_dev_app(self, app, env, venv, passthrough, **kwargs):
        self.actions.append(
            (
                "run_dev",
                app.app_name,
                app.test_mode,
                venv,
                passthrough,
                env,
            )
        )
        return full_options({"run_dev_state": app.app_name, "env": env}, kwargs)


@pytest.fixture
def dev_command(dummy_console, tmp_path):
    return DummyDevCommand(
        console=dummy_console,
        base_path=tmp_path,
    )


def test_no_args_one_app(dev_command, first_app):
    """If there is one app, dev starts that app by default."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }
    # Simulate that the venv already exists (installed app)
    first_venv = dev_command.simulate_existing_venv("first")

    # Configure no command line options
    options, _ = dev_command.parse_options([])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired
        ("virtual-environment", "first", True, False),
        # Run the first app devly
        (
            "run_dev",
            "first",
            False,
            first_venv,
            [],
            dev_command.env,
        ),
    ]

    # Environment was not recreated
    assert not first_venv.created


def test_no_args_two_apps(dev_command, first_app, second_app, monkeypatch):
    """If there are two apps and input is enabled, the user is prompted to pick one."""
    # Add two apps
    dev_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Interactive mode
    dev_command.console.input_enabled = True

    # Fake the user selecting "second"
    def fake_selection_question(**kwargs):
        return "second"

    monkeypatch.setattr(
        dev_command.console,
        "selection_question",
        fake_selection_question,
    )

    # No flags on the command line
    options, _ = dev_command.parse_options([])

    # This should follow the multi-app selection path and run without error.
    dev_command(**options)

    # A venv was created for the second app
    second_venv = dev_command.venvs[("second", True)]

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "second"),
        # App tools are verified for app
        ("verify-app-tools", "second"),
        # Virtual environment context is acquired
        ("virtual-environment", "second", True, False),
        # Requirements are installed into the venv
        ("dev_requirements", "second", second_venv),
        # Run the first app devly
        (
            "run_dev",
            "second",
            False,
            second_venv,
            [],
            dev_command.env,
        ),
    ]

    # Environment for second app was created
    assert second_venv.created


def test_no_args_two_apps_non_interactive(dev_command, first_app, second_app):
    """If there are two apps and --no-input is provided, an error is raised."""
    # Add two apps
    dev_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Simulate --no-input on the command line
    options, _ = dev_command.parse_options(["--no-input"])

    # In non-interactive mode, invoking the run command raises an error
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Project specifies more than one application",
    ):
        dev_command(**options)

    # Finalization will not occur
    assert dev_command.actions == []


def test_with_arg_one_app(dev_command, first_app):
    """If there is one app, and a -a argument, dev starts that app."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }

    # Simulate that the venv already exists (installed app)
    first_venv = dev_command.simulate_existing_venv("first")

    # Configure a -a command line option
    options, _ = dev_command.parse_options(["-a", "first"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired
        ("virtual-environment", "first", True, False),
        # Run the first app devly
        (
            "run_dev",
            "first",
            False,
            first_venv,
            [],
            dev_command.env,
        ),
    ]

    # Environment was not recreated
    assert not first_venv.created


def test_with_arg_two_apps(dev_command, first_app, second_app):
    """If there are multiple apps, the --app argument starts app nominated."""
    # Add two apps
    dev_command.apps = {
        "first": first_app,
        "second": second_app,
    }
    # Simulate that the venvs already exist (installed apps)
    first_venv = dev_command.simulate_existing_venv("first")
    second_venv = dev_command.simulate_existing_venv("second")

    # Configure a --app command line option
    options, _ = dev_command.parse_options(["--app", "second"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "second"),
        # App tools are verified for app
        ("verify-app-tools", "second"),
        # Virtual environment context is acquired
        ("virtual-environment", "second", True, False),
        # Run the second app devly
        (
            "run_dev",
            "second",
            False,
            second_venv,
            [],
            dev_command.env,
        ),
    ]

    # Neither environment was recreated
    assert not first_venv.created
    assert not second_venv.created


def test_create_venv(dev_command, first_app):
    """If no venv exists, it is created before starting."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }

    # Configure no command line options
    options, _ = dev_command.parse_options([])

    # Run the run command
    dev_command(**options)

    # An isolated venv was created
    first_venv = dev_command.venvs[("first", True)]

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired
        ("virtual-environment", "first", True, False),
        # Requirements are installed into the new venv
        ("dev_requirements", "first", first_venv),
        # Run the first app devly
        (
            "run_dev",
            "first",
            False,
            first_venv,
            [],
            dev_command.env,
        ),
    ]

    # Environment was created on first use
    assert first_venv.created

    # No clean calls were made.
    first_venv.clean.assert_not_called()


def test_create_venv_requirements_failure(dev_command, first_app):
    """If requirements can't be installed, the venv is cleaned up."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }

    # Mock a failure in the installation of development requirements
    dev_command.install_dev_requirements = mock.MagicMock(
        side_effect=RequirementsInstallError()
    )

    # Configure no command line options
    options, _ = dev_command.parse_options([])

    # Run the run command; it will raise an error
    with pytest.raises(
        RequirementsInstallError,
        match=r"Unable to install requirements\.",
    ):
        dev_command(**options)

    # An isolated venv was created
    first_venv = dev_command.venvs[("first", True)]

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired
        ("virtual-environment", "first", True, False),
        # No further actions are recorded due to the requirements failure.
    ]

    # Environment was created on first use
    assert first_venv.created

    # Clean was called to destroy the failed environment
    first_venv.clean.assert_called_once_with()


def test_non_isolated(dev_command, first_app):
    """If run in non-isolated mode, a no-op venv is created on first run."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }

    # Configure no command line options
    options, _ = dev_command.parse_options(["--no-isolation"])

    # Run the run command
    dev_command(**options)

    # A non-isolated environment now exists
    first_venv = dev_command.venvs[("first", False)]

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # No-op virtual environment context is acquired
        ("virtual-environment", "first", False, False),
        # Requirements are installed into the venv
        ("dev_requirements", "first", first_venv),
        # Run the first app devly
        (
            "run_dev",
            "first",
            False,
            first_venv,
            [],
            dev_command.env,
        ),
    ]

    # Non-isolated environment was created
    assert first_venv.created


def test_bad_app_reference(dev_command, first_app, second_app):
    """If the command line argument refers to an app that doesn't exist, raise an
    error."""
    # Add two apps
    dev_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure a --app command line option
    options, _ = dev_command.parse_options(["--app", "does-not-exist"])

    # Invoking the run command raises an error
    with pytest.raises(BriefcaseCommandError):
        dev_command(**options)

    # Finalization will not occur.
    assert dev_command.actions == []


def test_update_requirements(dev_command, first_app):
    """The dev command can request that the app is updated first."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }
    # Simulate that the venv already exists (installed app)
    first_venv = dev_command.simulate_existing_venv("first")

    # Configure a requirements update
    options, _ = dev_command.parse_options(["-r"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired with recreate=True
        ("virtual-environment", "first", True, True),
        # Requirements were installed again
        ("dev_requirements", "first", first_venv),
        # Then, it will be started
        (
            "run_dev",
            "first",
            False,
            first_venv,
            [],
            dev_command.env,
        ),
    ]

    # Environment was recreated
    assert first_venv.created

    # A clean call will have been made internally, but that's not tracked
    # by the mock. No *additional* clean calls were made.
    first_venv.clean.assert_not_called()


def test_install_requirements_failure(dev_command, first_app):
    """If the installation of requirements fails, the dev environment is cleaned."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }
    # Simulate that the venv already exists (installed app)
    first_venv = dev_command.simulate_existing_venv("first")

    # Mock a failure in the installation of development requirements
    dev_command.install_dev_requirements = mock.MagicMock(
        side_effect=RequirementsInstallError()
    )

    # Configure a requirements update
    options, _ = dev_command.parse_options(["-r"])

    # Run the run command; it will raise an error
    with pytest.raises(
        RequirementsInstallError,
        match=r"Unable to install requirements\.",
    ):
        dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired with recreate=True
        ("virtual-environment", "first", True, True),
        # But installing requirements fails.
    ]

    # Environment was recreated
    assert first_venv.created

    # Clean was called as a result of the failure.
    first_venv.clean.assert_called_once_with()


def test_run_uninstalled(dev_command, first_app_uninstalled):
    """The dev command will install first if the app hasn't been installed."""
    # Add a single app
    dev_command.apps = {
        "first": first_app_uninstalled,
    }

    # Configure no command line options
    options, _ = dev_command.parse_options([])

    # Run the run command
    dev_command(**options)

    # An isolated venv now exists
    first_venv = dev_command.venvs[("first", True)]

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired
        ("virtual-environment", "first", True, False),
        # The app will be installed
        ("dev_requirements", "first", first_venv),
        # Then, it will be started
        (
            "run_dev",
            "first",
            False,
            first_venv,
            [],
            dev_command.env,
        ),
    ]

    # Isolated environment was created
    assert first_venv.created


def test_update_uninstalled(dev_command, first_app_uninstalled):
    """A request to update requirements is redundant if the app hasn't been
    installed."""
    # Add a single app
    dev_command.apps = {
        "first": first_app_uninstalled,
    }

    # Configure a requirements update
    options, _ = dev_command.parse_options(["-r"])

    # Run the run command
    dev_command(**options)

    # An isolated venv now exists
    first_venv = dev_command.venvs[("first", True)]

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired with recreate=True
        ("virtual-environment", "first", True, True),
        # An update was requested
        ("dev_requirements", "first", first_venv),
        # Then, it will be started
        (
            "run_dev",
            "first",
            False,
            first_venv,
            [],
            dev_command.env,
        ),
    ]

    # An isolated environment was created
    assert first_venv.created


def test_no_run(dev_command, first_app_uninstalled):
    """Install requirements without running the app."""
    # Add a single app
    dev_command.apps = {
        "first": first_app_uninstalled,
    }

    # Configure an update without run
    options, _ = dev_command.parse_options(["--no-run"])

    # Run the run command
    dev_command(**options)

    # An isolated environment now exists
    first_venv = dev_command.venvs[("first", True)]

    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired with recreate=True
        ("virtual-environment", "first", True, True),
        # An update was requested
        ("dev_requirements", "first", first_venv),
    ]

    # An isolated environment was created
    assert first_venv.created


def test_run_test(dev_command, first_app):
    """The test suite can be run in development mode."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }
    # Simulate that the venv already exists (installed app)
    first_venv = dev_command.simulate_existing_venv("first")

    # Configure the test option
    options, _ = dev_command.parse_options(["--test"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired
        ("virtual-environment", "first", True, False),
        # Then, it will be started
        (
            "run_dev",
            "first",
            True,
            first_venv,
            [],
            dev_command.env,
        ),
    ]

    # Environment was not recreated
    assert not first_venv.created


def test_run_test_uninstalled(dev_command, first_app_uninstalled):
    """The test suite can be run in development mode."""
    # Add a single app
    dev_command.apps = {
        "first": first_app_uninstalled,
    }

    # Configure the test option
    options, _ = dev_command.parse_options(["--test"])

    # Run the run command
    dev_command(**options)

    # An isolated venv was created
    first_venv = dev_command.venvs[("first", True)]

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Virtual environment context is acquired
        ("virtual-environment", "first", True, False),
        # Development requirements will be installed
        ("dev_requirements", "first", first_venv),
        # Then, it will be started
        (
            "run_dev",
            "first",
            True,
            first_venv,
            [],
            dev_command.env,
        ),
    ]

    # Isolaved environment was created
    assert first_venv.created



import itertools
from unittest import mock
from unittest.mock import MagicMock


def test_all_config_files_are_read(base_command, mock_git):
    """All git config files are read (system, global, user, repo)."""
    base_command.tools.git = mock_git
    mock_git.config.get_config_path.side_effect = ["file1", "file2", "file3"]

    base_command.get_git_config_value("test-section", "test-option")

    assert mock_git.config.get_config_path.call_args_list == [
        mock.call("system"),
        mock.call("global"),
        mock.call("user"),
    ]
    expected_config_files = ["file1", "file2", "file3", ".git/config"]
    mock_git.config.GitConfigParser.assert_called_once_with(expected_config_files)


def test_config_values_are_parsed(base_command, tmp_path, monkeypatch):
    """If the requested value exists in one of the config files, it shall be
    returned."""
    import git

    # use 'real' gitpython library (no mock)
    base_command.tools.git = git

    # mock `git.config.get_config_path` to always provide the same three local files
    mock_config_paths = ["missing-file-1", "config-1", "missing-file-2"]
    git.config.get_config_path = MagicMock()
    git.config.get_config_path.side_effect = itertools.cycle(mock_config_paths)

    # create local two config files
    monkeypatch.chdir(tmp_path)
    (tmp_path / "config-1").write_text("[user]\n\tname = Some User\n", encoding="utf-8")
    (tmp_path / ".git").mkdir()
    (tmp_path / ".git" / "config").write_text(
        "[user]\n\temail = my@email.com\n", encoding="utf-8"
    )

    # expect values are parsed from all existing config files
    assert base_command.get_git_config_value("user", "name") == "Some User"
    assert base_command.get_git_config_value("user", "email") == "my@email.com"

    # expect that missing sections and options are handled
    assert base_command.get_git_config_value("user", "something") is None
    assert base_command.get_git_config_value("something", "something") is None



import sys
from pathlib import Path

import pytest

from briefcase.integrations.virtual_environment import VenvContext


@pytest.mark.parametrize(
    "empty_args",
    [
        [],
        (),
        None,
    ],
)
def test_rewrite_head_empty(venv_context: VenvContext, empty_args):
    """rewrite_head returns empty inputs unchanged."""
    result = venv_context._rewrite_head(empty_args)
    assert result == empty_args


@pytest.mark.parametrize(
    ("args", "expected_suffix"),
    [
        (
            [sys.executable],
            [],
        ),
        (
            [sys.executable, "-m", "pip", "install", "package"],
            ["-m", "pip", "install", "package"],
        ),
        (
            [Path(sys.executable), "-c", "print('hello')"],
            ["-c", "print('hello')"],
        ),
    ],
)
def test_rewrite_head_system_python_replacement(
    venv_context: VenvContext,
    args,
    expected_suffix,
):
    """rewrite_head replaces sys.executable with venv executable."""
    result = venv_context._rewrite_head(args)

    expected = [venv_context.executable, *expected_suffix]
    assert result == expected


@pytest.mark.parametrize(
    "args",
    [
        ["/usr/bin/python3"],
        ["pip", "install", "package"],
        ("python", "-c", "import sys"),
    ],
)
def test_rewrite_head_no_replacement(venv_context: VenvContext, args):
    """rewrite_head preserves non-matching commands and converts to list."""
    result = venv_context._rewrite_head(args)
    expected = list(args)
    assert result == expected
    assert isinstance(result, list)


@pytest.mark.skipif(sys.platform == "win32", reason="Unix specific test")
def test_rewrite_head_case_insensitive_match(venv_context):
    """rewrite_head handles case-insensitive matching via normcase."""
    case_variant = sys.executable

    args = [case_variant, "-V"]
    result = venv_context._rewrite_head(args)

    expected = [venv_context.executable, "-V"]
    assert result == expected


@pytest.mark.skipif(sys.platform != "win32", reason="Windows specific test")
def test_rewrite_head_case_insensitive_match_nt(venv_context):
    """rewrite_head handles case-insensitive matching."""
    case_variant = (
        sys.executable.upper() if sys.executable.islower() else sys.executable.lower()
    )
    args = [case_variant, "-V"]
    result = venv_context._rewrite_head(args)

    expected = [venv_context.executable, "-V"]
    assert result == expected



from unittest.mock import MagicMock


def test_popen_passes_through_to_subprocess(noop_context, mock_noop_subprocess_setup):
    """Popen passes all arguments to tools.subprocess.Popen."""
    mocks = mock_noop_subprocess_setup
    mock_popen = MagicMock()
    mocks["subprocess"].Popen.return_value = mock_popen

    args = ["python", "-c", "print('hello')"]
    kwargs = {"stdout": "PIPE", "stderr": "PIPE", "env": {"TEST": "value"}}

    result = noop_context.Popen(args, **kwargs)

    mocks["subprocess"].Popen.assert_called_once_with(args, **kwargs)
    assert result is mock_popen


def test_popen_with_no_kwargs(noop_context, mock_noop_subprocess_setup):
    """Popen works with just args, no kwargs."""
    mocks = mock_noop_subprocess_setup
    mock_popen = MagicMock()
    mocks["subprocess"].Popen.return_value = mock_popen

    args = ["python", "-V"]
    result = noop_context.Popen(args)

    mocks["subprocess"].Popen.assert_called_once_with(args)
    assert result is mock_popen


def test_popen_preserves_original_kwargs(noop_context, mock_noop_subprocess_setup):
    """Popen does not modify the original kwargs dict."""
    mocks = mock_noop_subprocess_setup
    mock_popen = MagicMock()
    mocks["subprocess"].Popen.return_value = mock_popen

    original_kwargs = {"stdout": "PIPE", "env": {"CUSTOM": "value"}}
    kwargs_copy = original_kwargs.copy()

    result = noop_context.Popen(["test"], **original_kwargs)

    assert original_kwargs == kwargs_copy
    mocks["subprocess"].Popen.assert_called_once_with(["test"], **original_kwargs)
    assert result is mock_popen



from subprocess import CalledProcessError
from unittest import mock
from zipfile import ZipFile

import pytest

import briefcase.platforms.windows
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.subprocess import Subprocess
from briefcase.integrations.windows_sdk import WindowsSDK
from briefcase.integrations.wix import WiX
from briefcase.platforms.windows.app import WindowsAppPackageCommand

from ....integrations.wix.conftest import WIX_EXE_PATH, WIX_UI_PATH
from ....utils import create_file


@pytest.fixture
def package_command(dummy_console, tmp_path):
    command = WindowsAppPackageCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.host_os = "Windows"
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command.tools.wix = WiX(command.tools, wix_home=tmp_path / "wix")
    command.tools.windows_sdk = WindowsSDK(
        tools=command.tools,
        root_path=tmp_path / "windows_sdk",
        version="81.2.1.0",
        arch="groovy",
    )
    return command


@pytest.fixture
def package_command_with_files(package_command, tmp_path):
    # Build the paths for the source and the distribution folders:
    src_path = tmp_path / "base_path/build/first-app/windows/app/src"
    dist_path = tmp_path / "base_path/dist"
    src_path.mkdir(parents=True)
    dist_path.mkdir(parents=True)

    # Mock some typical folders and files in the src folder:
    files = (
        src_path / "First App.exe",
        src_path / "python.exe",
        src_path / "python3.dll",
        src_path / "vcruntime140.dll",
        src_path / "app/first-app/app.py",
        src_path / "app/first-app/resources/__init__.py",
        src_path / "app/first-app-0.0.1.dist-info/top_level.txt",
        src_path / "app_packages/clr.py",
        src_path / "app_packages/toga_winforms/command.py",
    )
    for file in files:
        create_file(file, "")

    return package_command


def test_package_formats(package_command):
    """Packaging formats are as expected."""
    assert package_command.packaging_formats == ["msi", "zip"]
    assert package_command.default_packaging_format == "msi"


def test_verify(package_command, monkeypatch):
    """Verifying on Windows creates a WiX wrapper."""
    # prime Command to _not_ need Windows SDK
    package_command._windows_sdk_needed = False

    mock_wix_verify = mock.MagicMock(wraps=WiX.verify)
    monkeypatch.setattr(
        briefcase.platforms.windows.WiX,
        "verify",
        mock_wix_verify,
    )

    package_command.verify_tools()

    # WiX tool was verified
    mock_wix_verify.assert_called_once_with(tools=package_command.tools)
    assert isinstance(package_command.tools.wix, WiX)


def test_verify_with_signing(package_command, monkeypatch):
    """Verifying on Windows creates WiX and WindowsSDK wrappers when code signing."""
    # prime Command to need Windows SDK
    package_command._windows_sdk_needed = True

    mock_windows_sdk_verify = mock.MagicMock(wraps=WindowsSDK.verify)
    monkeypatch.setattr(
        briefcase.platforms.windows.WindowsSDK,
        "verify",
        mock_windows_sdk_verify,
    )

    mock_wix_verify = mock.MagicMock(wraps=WiX.verify)
    monkeypatch.setattr(
        briefcase.platforms.windows.WiX,
        "verify",
        mock_wix_verify,
    )

    package_command.verify_tools()

    # WiX tool was verified
    mock_wix_verify.assert_called_once_with(tools=package_command.tools)
    assert isinstance(package_command.tools.wix, WiX)
    # WindowsSDK tool was verified
    mock_windows_sdk_verify.assert_called_once_with(tools=package_command.tools)
    assert isinstance(package_command.tools.windows_sdk, WindowsSDK)


@pytest.mark.parametrize(
    ("cli_args", "signing_options", "is_sdk_needed"),
    [
        ([], {}, False),
        (["--adhoc-sign"], {"adhoc_sign": True}, False),
        (["--file-digest", "sha2000"], {"file_digest": "sha2000"}, False),
        (["-i", "asdf"], {"identity": "asdf"}, True),
        (["--identity", "asdf"], {"identity": "asdf"}, True),
        (
            [
                "-i",
                "asdf",
                "--file-digest",
                "sha42",
                "--use-local-machine-stores",
                "--cert-store",
                "mystore",
                "--timestamp-url",
                "http://freetimestamps.com",
                "--timestamp-digest",
                "sha56",
            ],
            {
                "identity": "asdf",
                "file_digest": "sha42",
                "use_local_machine": True,
                "cert_store": "mystore",
                "timestamp_url": "http://freetimestamps.com",
                "timestamp_digest": "sha56",
            },
            True,
        ),
        (
            [
                "-i",
                "asdf",
                "--file-digest",
                "sha42",
                "--cert-store",
                "mystore",
                "--timestamp-url",
                "http://freetimestamps.com",
                "--timestamp-digest",
                "sha56",
            ],
            {
                "identity": "asdf",
                "file_digest": "sha42",
                "cert_store": "mystore",
                "timestamp_url": "http://freetimestamps.com",
                "timestamp_digest": "sha56",
            },
            True,
        ),
    ],
)
def test_parse_options(package_command, cli_args, signing_options, is_sdk_needed):
    """Command line arguments are parsed as expected; Windows SDK is required if an
    identity is specified."""
    default_options = {
        "identity": None,
        "file_digest": "sha256",
        "use_local_machine": False,
        "cert_store": "My",
        "timestamp_url": "http://timestamp.digicert.com",
        "timestamp_digest": "sha256",
        "adhoc_sign": False,
        "packaging_format": "msi",
        "update": False,
    }
    expected_options = {**default_options, **signing_options}

    options, overrides = package_command.parse_options(extra=cli_args)

    assert options == expected_options
    assert overrides == {}
    assert package_command._windows_sdk_needed is is_sdk_needed


@pytest.mark.parametrize(
    ("kwargs", "external"),
    [
        # Default behavior (adhoc signing, internal app)
        ({}, False),
        # External app mode makes no difference when packaging an MSI, because it's
        # entirely implemented in `briefcase create`.
        ({}, True),
        # Explicit adhoc signing
        ({"adhoc_sign": True}, False),
    ],
)
def test_package_msi(
    package_command, first_app_config, external_first_app, tmp_path, kwargs, external
):
    """A Windows app can be packaged as an MSI."""

    package_command.package_app(
        external_first_app if external else first_app_config,
        **kwargs,
    )

    assert package_command.tools.subprocess.run.mock_calls == [
        # Compile MSI
        mock.call(
            [
                tmp_path / "wix" / WIX_EXE_PATH,
                "build",
                "-ext",
                tmp_path / "wix" / WIX_UI_PATH,
                "-arch",
                "x64",
                "first-app.wxs",
                "-loc",
                "unicode.wxl",
                "-pdbtype",
                "none",
                "-o",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
    ]


@pytest.mark.parametrize(
    "kwargs",
    [
        {},  # Default behavior (adhoc signing)
        {"adhoc_sign": True},  # Explicit adhoc signing
    ],
)
def test_package_zip(package_command_with_files, first_app_config, kwargs, tmp_path):
    """A Windows app can be packaged as a zip file."""

    first_app_config.packaging_format = "zip"
    package_command_with_files.package_app(first_app_config, **kwargs)

    # No signing was performed
    assert package_command_with_files.tools.subprocess.run.mock_calls == []

    archive_file = tmp_path / "base_path/dist/First App-0.0.1.zip"
    source_folders_and_files = (
        "app/",
        "app/first-app/",
        "app/first-app/resources/",
        "app/first-app-0.0.1.dist-info/",
        "app_packages/",
        "app_packages/toga_winforms/",
        "First App.exe",
        "python.exe",
        "python3.dll",
        "vcruntime140.dll",
        "app/first-app/app.py",
        "app/first-app/resources/__init__.py",
        "app/first-app-0.0.1.dist-info/top_level.txt",
        "app_packages/clr.py",
        "app_packages/toga_winforms/command.py",
    )

    # The zip file exists
    assert archive_file.exists()

    # Check content of zip file
    with ZipFile(archive_file) as archive:
        root = "First App-0.0.1/"
        # All folders and files in zip are from source
        for name in archive.namelist():
            # name.removeprefix(f'{root}') will only work in Python > 3.8
            assert name[len(root) :] in source_folders_and_files
        # All files from source are in zip
        for file in source_folders_and_files:
            assert f"{root}{file}" in archive.namelist()


@pytest.mark.parametrize(
    ("use_local_machine", "additional_args"),
    [(False, []), (True, ["-sm"])],
)
def test_package_msi_with_codesigning(
    package_command,
    first_app_config,
    tmp_path,
    use_local_machine,
    additional_args,
):
    """A Windows app can be packaged as an MSI and code signed."""

    package_command.package_app(
        first_app_config,
        identity="80ee4c3321122916f5637522451993c2a0a4a56a",
        file_digest="sha42",
        use_local_machine=use_local_machine,
        cert_store="mystore",
        timestamp_url="http://freetimestamps.com",
        timestamp_digest="sha56",
    )

    assert package_command.tools.subprocess.run.mock_calls == [
        # Codesign app exe
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
                *additional_args,
                tmp_path
                / "base_path"
                / "build"
                / "first-app"
                / "windows"
                / "app"
                / "src"
                / "First App.exe",
            ],
            check=True,
        ),
        # Compile MSI
        mock.call(
            [
                tmp_path / "wix" / WIX_EXE_PATH,
                "build",
                "-ext",
                tmp_path / "wix" / WIX_UI_PATH,
                "-arch",
                "x64",
                "first-app.wxs",
                "-loc",
                "unicode.wxl",
                "-pdbtype",
                "none",
                "-o",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Codesign app MSI
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
                *additional_args,
                tmp_path / "base_path/dist/First App-0.0.1.msi",
            ],
            check=True,
        ),
    ]


@pytest.mark.parametrize(
    ("use_local_machine", "additional_args"),
    [(False, []), (True, ["-sm"])],
)
def test_package_zip_with_codesigning(
    package_command_with_files,
    first_app_config,
    tmp_path,
    use_local_machine,
    additional_args,
):
    """In a ZIP package, only the binary will be code signed."""

    first_app_config.packaging_format = "zip"

    package_command_with_files.package_app(
        first_app_config,
        identity="80ee4c3321122916f5637522451993c2a0a4a56a",
        file_digest="sha42",
        use_local_machine=use_local_machine,
        cert_store="mystore",
        timestamp_url="http://freetimestamps.com",
        timestamp_digest="sha56",
    )

    assert package_command_with_files.tools.subprocess.run.mock_calls == [
        # Codesign app exe
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
                *additional_args,
                tmp_path / "base_path/build/first-app/windows/app/src/First App.exe",
            ],
            check=True,
        ),
    ]


def test_package_msi_invalid_identity(package_command, first_app_config):
    """Codesigning fails, along with packaging, if the identity is invalid."""
    with pytest.raises(
        BriefcaseCommandError,
        match="Codesigning identify 'asdf' must be a certificate SHA-1 thumbprint",
    ):
        package_command.package_app(first_app_config, identity="asdf")


def test_package_msi_failed_sign_app(package_command, first_app_config, tmp_path):
    """An error is raised if signtool fails for the app exe."""
    # Mock a failure in the call to signtool.exe
    package_command.tools.subprocess.run.side_effect = [
        CalledProcessError(cmd=["signtool.exe"], returncode=1),
    ]

    with pytest.raises(BriefcaseCommandError, match=r"Unable to sign "):
        package_command.package_app(
            first_app_config,
            identity="80ee4c3321122916f5637522451993c2a0a4a56a",
            file_digest="sha42",
            cert_store="mystore",
            timestamp_url="http://freetimestamps.com",
            timestamp_digest="sha56",
        )

    assert package_command.tools.subprocess.run.mock_calls == [
        # Codesign app exe
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
                tmp_path / "base_path/build/first-app/windows/app/src/First App.exe",
            ],
            check=True,
        )
    ]


def test_package_msi_failed_compile(package_command, first_app_config, tmp_path):
    """An error is raised if compilation failed."""
    # Mock a failure in the call to wix.exe
    package_command.tools.subprocess.run.side_effect = [
        CalledProcessError(cmd=["wix.exe"], returncode=1),
    ]

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to package app first-app.",
    ):
        package_command.package_app(first_app_config)

    assert package_command.tools.subprocess.run.mock_calls == [
        # Compile MSI
        mock.call(
            [
                tmp_path / "wix" / WIX_EXE_PATH,
                "build",
                "-ext",
                tmp_path / "wix" / WIX_UI_PATH,
                "-arch",
                "x64",
                "first-app.wxs",
                "-loc",
                "unicode.wxl",
                "-pdbtype",
                "none",
                "-o",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
    ]


def test_package_msi_failed_signing_msi(package_command, first_app_config, tmp_path):
    """An error is raised if signtool fails for the app MSI."""
    # Mock a failure in the call to signtool.exe
    package_command.tools.subprocess.run.side_effect = [
        None,  # signtool.exe
        None,  # wix.exe
        CalledProcessError(cmd=["signtool.exe"], returncode=1),
    ]

    with pytest.raises(BriefcaseCommandError, match=r"Unable to sign "):
        package_command.package_app(
            first_app_config,
            identity="80ee4c3321122916f5637522451993c2a0a4a56a",
            file_digest="sha42",
            cert_store="mystore",
            timestamp_url="http://freetimestamps.com",
            timestamp_digest="sha56",
        )

    assert package_command.tools.subprocess.run.mock_calls == [
        # Codesign app exe
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
                tmp_path / "base_path/build/first-app/windows/app/src/First App.exe",
            ],
            check=True,
        ),
        # Compile MSI
        mock.call(
            [
                tmp_path / "wix" / WIX_EXE_PATH,
                "build",
                "-ext",
                tmp_path / "wix" / WIX_UI_PATH,
                "-arch",
                "x64",
                "first-app.wxs",
                "-loc",
                "unicode.wxl",
                "-pdbtype",
                "none",
                "-o",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Codesign app MSI
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
            ],
            check=True,
        ),
    ]


def test_external_package_zip(
    package_command_with_files,
    external_first_app,
    tmp_path,
):
    """In a ZIP package, only the binary will be code signed."""

    external_first_app.packaging_format = "zip"

    package_command_with_files.package_app(
        external_first_app,
        identity="80ee4c3321122916f5637522451993c2a0a4a56a",
        file_digest="sha42",
        use_local_machine=False,
        cert_store="mystore",
        timestamp_url="http://freetimestamps.com",
        timestamp_digest="sha56",
    )

    assert package_command_with_files.tools.subprocess.run.mock_calls == [
        # Codesign app exe
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
                tmp_path / "base_path/external/src/internal/app.exe",
            ],
            check=True,
        ),
    ]



import pytest

from tests.integrations.virtual_environment.utils import (
    ENVIRONMENT_TEST_PARAMS,
    assert_environment_handling,
)


@pytest.mark.parametrize(("env_override", "other_kwargs"), ENVIRONMENT_TEST_PARAMS)
def test_popen_environment_handling(
    venv_context,
    mock_subprocess_setup,
    env_override,
    other_kwargs,
):
    """Popen merges venv with overrides and passes all kwargs to subprocess."""
    mocks = mock_subprocess_setup

    kwargs = other_kwargs.copy()
    if env_override is not None:
        kwargs["env"] = env_override

    original_kwargs = kwargs.copy()

    result = venv_context.Popen(["test"], **kwargs)

    assert kwargs == original_kwargs
    mocks["rewrite_head"].assert_called_once_with(["test"])

    # The venv's full_env should be called with any provided env override, and the
    # resulting environment should be passed to subprocess.check_output along with other
    # kwargs unchanged.
    assert_environment_handling(
        mock_full_env=mocks["full_env"],
        env_override=env_override,
        mock_method=mocks["subprocess"].Popen,
        method_args=["rewritten", "args"],
        other_kwargs=other_kwargs,
    )

    assert result is mocks["popen_instance"]


def test_popen_no_args(venv_context, mock_subprocess_setup):
    """Popen handles empty args list."""
    mocks = mock_subprocess_setup
    mocks["rewrite_head"].return_value = []

    result = venv_context.Popen([])

    mocks["rewrite_head"].assert_called_once_with([])
    mocks["full_env"].assert_called_once_with(None)
    mocks["subprocess"].Popen.assert_called_once_with([], env={"FULL": "env"})

    assert result is mocks["popen_instance"]


def test_popen_complex_args_rewriting(venv_context, mock_subprocess_setup):
    """Popen handles complex argument rewriting."""
    mocks = mock_subprocess_setup
    mocks["rewrite_head"].return_value = [
        "/venv/bin/python",
        "-m",
        "pip",
        "install",
        "package",
    ]

    original_args = ["/usr/bin/python", "-m", "pip", "install", "package"]
    result = venv_context.Popen(original_args)

    mocks["rewrite_head"].assert_called_once_with(original_args)
    mocks["full_env"].assert_called_once_with(None)
    mocks["subprocess"].Popen.assert_called_once_with(
        ["/venv/bin/python", "-m", "pip", "install", "package"], env={"FULL": "env"}
    )

    assert result is mocks["popen_instance"]



from briefcase.integrations.virtual_environment import VenvContext

from ...utils import create_file


def test_clean(mock_tools, venv_path):
    """Clean the venv at the specified path."""
    context = VenvContext(mock_tools, venv_path)

    # Create a dummy venv configuration
    create_file(venv_path / "pyvenv.cfg", "VENV CONFIG")

    # venv exists before clean
    assert context.exists()

    # Try to clean the venv
    context.clean()

    # venv doesn't exist before clean
    assert not context.exists()


def test_clean_non_existent(mock_tools, venv_path):
    """It's possible to clean a venv that doesn't exist."""

    context = VenvContext(mock_tools, venv_path)

    # venv doesn't exist before clean
    assert not context.exists()

    # Try to clean the venv
    context.clean()

    # venv doesn't exist before clean
    assert not context.exists()



from unittest.mock import PropertyMock

import pytest
from rich.console import ColorSystem


@pytest.mark.parametrize(
    ("no_color", "color_system", "is_enabled"),
    [
        (False, ColorSystem.TRUECOLOR, True),
        (False, None, False),
        (True, ColorSystem.TRUECOLOR, False),
        (True, None, False),
    ],
)
def test_is_color_enabled(
    raw_console,
    no_color,
    color_system,
    is_enabled,
    monkeypatch,
):
    """Color is enabled/disabled based on no_color and color_system."""
    raw_console._console_impl.no_color = no_color
    monkeypatch.setattr(
        type(raw_console._console_impl),
        "color_system",
        PropertyMock(return_value=color_system),
    )

    # confirm these values to make sure they didn't change somehow...
    assert raw_console._console_impl.no_color is no_color
    assert raw_console._console_impl.color_system is color_system

    assert raw_console.is_color_enabled is is_enabled



from briefcase.integrations.virtual_environment import NoOpVenvContext

from ...utils import create_file


def test_clean(mock_tools, venv_path):
    """Clean the venv at the specified path."""
    context = NoOpVenvContext(mock_tools, venv_path)

    # Create a venv marker file
    create_file(venv_path / "venv_path", "VENV PATH")

    # Try to clean the venv
    context.clean()

    # Marker file has been deleted
    assert not (venv_path / "venv_path").is_file()


def test_clean_non_existent(mock_tools, venv_path):
    """It's possible to clean a venv that doesn't exist."""

    context = NoOpVenvContext(mock_tools, venv_path)

    # Marker file doesn't exist
    assert not (venv_path / "venv_path").is_file()

    # Try to clean the venv
    context.clean()

    # Marker file still doesn't exist
    assert not (venv_path / "venv_path").is_file()



from briefcase.integrations.virtual_environment import VenvContext


def test_exists_when_venv_and_config_exists(mock_tools, venv_path):
    """A venv exists if both the directory and pyvenv.cfg file are present."""
    venv_path.mkdir()
    (venv_path / "pyvenv.cfg").touch()
    context = VenvContext(mock_tools, venv_path)
    assert context.exists() is True


def test_exists_when_venv_path_missing(mock_tools, tmp_path):
    """A venv does not exist if the directory is missing."""
    venv = tmp_path / "nonexistent_venv"
    context = VenvContext(mock_tools, venv)
    assert context.exists() is False


def test_exists_when_pyvenv_cfg_missing(mock_tools, venv_path):
    """A venv does not exist if the directory exists but pyvenv.cfg is missing."""
    venv_path.mkdir()
    context = VenvContext(mock_tools, venv_path)
    assert context.exists() is False



from unittest.mock import MagicMock

import pytest

from briefcase.commands.new import LICENSE_OPTIONS

from ...utils import PartialMatchString


@pytest.mark.parametrize("license_file_name", ["LICENSE", "LICENCE"])
@pytest.mark.parametrize(
    ("license_text", "license_id"),
    [
        ("MIT", "MIT"),
        ("MIT license", "MIT"),
        ("Permission is hereby granted, free of charge", "MIT"),
        ("Apache license", "Apache-2.0"),
        ("BSD", "BSD-3-Clause"),
        ("BSD license", "BSD-3-Clause"),
        # Includes some extra text to ensure it doesn't get caught as MIT because of
        # perMITted
        (
            (
                "Redistribution and use in source and binary forms, with or without"
                " modification, are permitted"
            ),
            "BSD-3-Clause",
        ),
        ("GPLv2", "GPL-2.0"),
        ("version 2 of the GNU General Public License", "GPL-2.0"),
        ("GPLv2+", "GPL-2.0+"),
        ("Free Software Foundation, either version 2 of the License", "GPL-2.0+"),
        ("GPLv3", "GPL-3.0"),
        ("version 3 of the GNU General Public License", "GPL-3.0"),
        ("GPLv3+", "GPL-3.0+"),
        ("either version 3 of the License", "GPL-3.0+"),
        ("Some text", "Other"),
    ],
)
def test_get_license_from_file(
    convert_command,
    license_text,
    license_id,
    license_file_name,
    monkeypatch,
):
    mock_selection_question = MagicMock()
    monkeypatch.setattr(
        convert_command.console, "selection_question", mock_selection_question
    )

    dummy_license_text = (
        "some leading text\neven_more_text" + license_text + "some_ending_text\n"
    )
    (convert_command.base_path / license_file_name).write_text(
        dummy_license_text, encoding="utf-8"
    )

    convert_command.input_license(None)

    mock_selection_question.assert_called_once_with(
        intro=PartialMatchString("Based on the license file"),
        description="Project License",
        options=LICENSE_OPTIONS,
        default=license_id,
        override_value=None,
    )


@pytest.mark.parametrize(
    ("license_text", "license_id"),
    [
        ("MIT", "MIT"),
        ("MIT license", "MIT"),
        ("Permission is hereby granted, free of charge", "MIT"),
        ("Apache license", "Apache-2.0"),
        ("BSD", "BSD-3-Clause"),
        ("BSD license", "BSD-3-Clause"),
        # Includes some extra text to ensure it doesn't get caught as MIT because of
        # perMITted
        (
            (
                "Redistribution and use in source and binary forms, with or without"
                " modification, are permitted"
            ),
            "BSD-3-Clause",
        ),
        ("GPLv2", "GPL-2.0"),
        ("version 2 of the GNU General Public License", "GPL-2.0"),
        ("GPLv2+", "GPL-2.0+"),
        ("Free Software Foundation, either version 2 of the License", "GPL-2.0+"),
        ("GPLv3", "GPL-3.0"),
        ("version 3 of the GNU General Public License", "GPL-3.0"),
        ("GPLv3+", "GPL-3.0+"),
        ("either version 3 of the License", "GPL-3.0+"),
        ("Some text", "Other"),
    ],
)
def test_get_license_from_pep621_license_file(
    convert_command,
    license_text,
    license_id,
    monkeypatch,
):
    mock_selection_question = MagicMock()
    monkeypatch.setattr(
        convert_command.console, "selection_question", mock_selection_question
    )

    dummy_license_text = (
        "some leading text\neven_more_text" + license_text + "some_ending_text\n"
    )
    (convert_command.base_path / "LICENSE.txt").write_text(
        dummy_license_text, encoding="utf-8"
    )
    (convert_command.base_path / "pyproject.toml").write_text(
        '[project]\nlicense = {file = "LICENSE.txt"}', encoding="utf-8"
    )

    convert_command.input_license(None)

    mock_selection_question.assert_called_once_with(
        intro=PartialMatchString("Based on the license file"),
        description="Project License",
        options=LICENSE_OPTIONS,
        default=license_id,
        override_value=None,
    )


@pytest.mark.parametrize(
    ("license_text", "license_id"),
    [
        ("MIT", "MIT"),
        ("MIT license", "MIT"),
        ("BSD", "BSD-3-Clause"),
        ("BSD license", "BSD-3-Clause"),
        ("GPLv2", "GPL-2.0"),
        ("GPLv2+", "GPL-2.0+"),
        ("GPLv3", "GPL-3.0"),
        ("GPLv3+", "GPL-3.0+"),
        ("Some text", "Other"),
    ],
)
def test_get_license_from_pyproject(
    convert_command,
    license_text,
    license_id,
    monkeypatch,
):
    mock_selection_question = MagicMock()
    monkeypatch.setattr(
        convert_command.console, "selection_question", mock_selection_question
    )
    (convert_command.base_path / "pyproject.toml").write_text(
        f'[project]\nlicense = {{text = "{license_text}"}}', encoding="utf-8"
    )

    convert_command.input_license(None)

    mock_selection_question.assert_called_once_with(
        intro=PartialMatchString("Based on the PEP621 formatted pyproject.toml"),
        description="Project License",
        options=LICENSE_OPTIONS,
        default=license_id,
        override_value=None,
    )


def test_no_license_hint(convert_command, monkeypatch):
    mock_selection_question = MagicMock()
    monkeypatch.setattr(
        convert_command.console, "selection_question", mock_selection_question
    )

    convert_command.input_license(None)

    assert mock_selection_question.call_count == 1
    assert mock_selection_question.call_args.kwargs["default"] is None
    assert "Based on" not in mock_selection_question.call_args.kwargs["intro"]


def test_override_is_used(convert_command):
    assert convert_command.input_license("Proprietary") == "Proprietary"



import shutil
import sys
from unittest import mock

if sys.version_info >= (3, 11):  # pragma: no-cover-if-lt-py311
    pass
else:  # pragma: no-cover-if-gte-py311
    pass

from zipfile import ZipFile

import pytest

from briefcase.exceptions import BriefcaseConfigError
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.web.static import StaticWebBuildCommand


@pytest.fixture
def build_command(dummy_console, tmp_path):
    command = StaticWebBuildCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command.tools.shutil = mock.MagicMock(spec_set=shutil)

    return command


def _mock_wheel(tmp_path, wheel_name, files):
    """Function to mock wheel files in tests."""
    wheel_path = tmp_path / wheel_name
    with ZipFile(wheel_path, "w") as wheel:
        for filename, content in files.items():
            wheel.writestr(filename, content)
    return wheel_path


def test_extract_pyscript_config(build_command, tmp_path):
    """Test function works correctly with both config.toml and pyscript.toml."""
    # Mock a wheel with files
    files = {
        "dependency/deploy/config.toml": """
implementation = "pyscript"

[pyscript]
version = "2024.10.1"
""",
        "dependency/deploy/pyscript.toml": """
existing-key-1 = "value-1"
existing-key-2 = 2
""",
    }

    wheel_path = _mock_wheel(tmp_path=tmp_path, wheel_name="dependency", files=files)

    # Run the function.
    pyscript_config, pyscript_version = build_command.extract_pyscript_config(
        [wheel_path]
    )

    # Check returns are correct and files were found.
    assert pyscript_config == {
        "existing-key-1": "value-1",
        "existing-key-2": 2,
    }
    assert pyscript_version == "2024.10.1"


def test_extract_pyscript_config_no_config(build_command, tmp_path):
    """If no config.toml supplied by wheels, functions returns a basic config."""
    # Mock a wheel without the needed files
    files = {"dependency/deploy/not-the-files-you-are-looking-for.toml": ""}

    wheel_path = _mock_wheel(tmp_path=tmp_path, wheel_name="dependency", files=files)

    # Run the function.
    pyscript_config, pyscript_version = build_command.extract_pyscript_config(
        [wheel_path]
    )

    # Check pyscript_config is empty and pyscript_version is the default.
    assert pyscript_config == {}
    assert pyscript_version == "2024.11.1"


def test_extract_pyscript_config_multiple_config(build_command, tmp_path):
    """Multiple config.toml supplied by wheels fails."""

    # Mock wheels that both contain config.toml
    file_set_1 = {
        "dependency/deploy/config.toml": """
implementation = "pyscript"

[pyscript]
version = "2024.10.1"
""",
    }

    file_set_2 = {
        "dependency/deploy/config.toml": """
implementation = "pyscript"

[pyscript]
version = "2024.10.1"
""",
    }

    wheel_path_1 = _mock_wheel(
        tmp_path=tmp_path, wheel_name="dependency", files=file_set_1
    )
    wheel_path_2 = _mock_wheel(
        tmp_path=tmp_path, wheel_name="dependency", files=file_set_2
    )

    # Run the function.
    with pytest.raises(
        BriefcaseConfigError,
        match=r"Only one deployment configuration file can be supplied.",
    ):
        build_command.extract_pyscript_config([wheel_path_1, wheel_path_2])


def test_extract_pyscript_config_no_implementation(build_command, tmp_path, capsys):
    """An app with no "implementation" value defaults to pyscript with a warning."""

    # Mock a wheel with the needed files
    files = {"dependency/deploy/config.toml": ""}

    wheel_path = _mock_wheel(tmp_path=tmp_path, wheel_name="dependency", files=files)

    pyscript_config, pyscript_version = build_command.extract_pyscript_config(
        [wheel_path]
    )

    # Capture output and assert warning present.
    captured = capsys.readouterr()
    assert "No web implementation specified. Defaulting to 'pyscript'." in captured.out

    # Check pyscript_config is empty and pyscript_version is the default.
    assert pyscript_config == {}
    assert pyscript_version == "2024.11.1"


def test_extract_pyscript_config_implementation_warning(
    build_command,
    tmp_path,
    capsys,
):
    """Briefcase raises a warning if "implementation" value is not pyscript."""

    files = {
        "dependency/deploy/config.toml": """
implementation = "something-else"
""",
    }

    wheel_path = _mock_wheel(tmp_path=tmp_path, wheel_name="dependency", files=files)

    pyscript_config, pyscript_version = build_command.extract_pyscript_config(
        [wheel_path]
    )

    # Capture output and assert warning present.
    captured = capsys.readouterr()
    assert (
        "At present, 'pyscript' is the only supported web implementation."
        in captured.out
    )

    # Check pyscript_config is empty and pyscript_version is the default.
    assert pyscript_config == {}
    assert pyscript_version == "2024.11.1"


def test_extract_pyscript_config_no_pyscript_toml(build_command, tmp_path, capsys):
    """If no pyscript.toml is supplied by a wheel, function returns a basic config."""

    # Create wheel with no pyscript.toml
    files = {
        "dependency/deploy/config.toml": """
implementation = "pyscript"
""",
    }

    wheel_path = _mock_wheel(tmp_path=tmp_path, wheel_name="dependency", files=files)

    pyscript_config, pyscript_version = build_command.extract_pyscript_config(
        [wheel_path]
    )
    # Capture output and assert console info is present.
    captured = capsys.readouterr()
    assert "Pyscript configuration file not found in" in captured.out

    # Check pyscript_config is empty and pyscript_version is the default.
    assert pyscript_config == {}
    assert pyscript_version == "2024.11.1"


def test_extract_pyscript_config_invalid_wheel_pyscript_toml(build_command, tmp_path):
    """A wheel with an invalid pyscript.toml raises an error."""

    # Mock a wheel with files
    files = {
        "dependency/deploy/config.toml": """
implementation = "pyscript"

[pyscript]
version = "2024.10.1"
""",
        "dependency/deploy/pyscript.toml": """
This is not valid toml.
""",
    }

    wheel_path = _mock_wheel(tmp_path=tmp_path, wheel_name="dependency", files=files)

    # Building the web app raises an error
    with pytest.raises(
        BriefcaseConfigError,
        match=r"Briefcase configuration error: pyscript.toml content isn't valid TOML: Expected",
    ):
        build_command.extract_pyscript_config([wheel_path])



import platform
from unittest import mock

import pytest

from briefcase.config import GlobalConfig
from briefcase.exceptions import BriefcaseConfigError, UnsupportedPythonVersion


def _get_global_config(requires_python):
    return GlobalConfig(
        project_name="pep621-requires-python-testing",
        version="0.0.1",
        bundle="com.example",
        requires_python=requires_python,
    )


def test_no_requires_python(base_command, my_app):
    """If requires-python isn't set, no verification is necessary."""

    base_command.global_config = _get_global_config(requires_python=None)
    base_command.verify_required_python(my_app)


@pytest.mark.parametrize(
    "requires_python",
    [
        "!= 3.2",
        ">= 3.2",
        "> 3.2",
        ">= {current}",
        "== {current}",
        "~= {current}",
        "<= {current}",
        "< 3.100",
    ],
)
def test_requires_python_met(base_command, my_app, requires_python):
    """Validation passes if requires-python specifies a version compatible with the
    running interpreter."""

    base_command.global_config = _get_global_config(
        requires_python.format(current=platform.python_version())
    )
    base_command.verify_required_python(my_app)


@pytest.mark.parametrize(
    "requires_python",
    [
        # Require a version higher than anything that can exist
        "> 3.100",
        ">= 3.100",
        # Require a version lower than anything that is supported
        "< 3.2",
        "<= 3.2",
        # Equality with a version that definitely isn't supported
        "== 2.0",
        "~= 2.0",
    ],
)
def test_requires_python_unmet(base_command, my_app, requires_python):
    """Validation fails if requires-python specifies a version incompatible with the
    running interpreter."""

    base_command.global_config = _get_global_config(requires_python)

    with pytest.raises(UnsupportedPythonVersion):
        base_command.verify_required_python(my_app)


def test_requires_python_invalid_specifier(base_command, my_app):
    """Validation fails if requires-python is not a valid specifier."""

    base_command.global_config = _get_global_config(requires_python="0")

    with pytest.raises(BriefcaseConfigError, match="Invalid requires-python"):
        base_command.verify_required_python(my_app)


def test_requires_python_met_specifier_set(monkeypatch, base_command, my_app):
    monkeypatch.setattr(platform, "python_version", mock.Mock(return_value="3.14.0"))

    base_command.global_config = _get_global_config(requires_python=">=3.14,<4.0")
    base_command.verify_required_python(my_app)


def test_requires_python_unmet_specifier_set(monkeypatch, base_command, my_app):
    monkeypatch.setattr(platform, "python_version", mock.Mock(return_value="3.13.0"))

    base_command.global_config = _get_global_config(requires_python=">=3.14,<4.0")
    with pytest.raises(UnsupportedPythonVersion):
        base_command.verify_required_python(my_app)


def test_requires_python_prerelease(monkeypatch, base_command, my_app):
    """Pre-release Python versions are included in matches."""
    monkeypatch.setattr(platform, "python_version", mock.Mock(return_value="3.14.0a0"))

    base_command.global_config = _get_global_config(requires_python=">=3.12")
    base_command.verify_required_python(my_app)



import pytest


@pytest.mark.parametrize(
    ("in_text", "out_text"),
    [
        (
            "There is nothing wrong with your television set.",
            "There is nothing wrong with your television set.",
        ),
        (
            (
                "There is nothing wrong with your television set.\n"
                "Do not attempt to adjust the picture."
            ),
            (
                "There is nothing wrong with your television set.\n"
                "Do not attempt to adjust the picture."
            ),
        ),
        (
            (
                "There is nothing\n\n\nwrong with your television set.\n\nDo not"
                " attempt to adjust the picture. We are controlling transmission. If we"
                " wish to make it louder, we will bring\nup the volume.\n"
            ),
            (
                "There is nothing\n\n\nwrong with your television set.\n\nDo not"
                " attempt to adjust the picture. We are controlling transmission. If"
                " we\nwish to make it louder, we will bring\nup the volume."
            ),
        ),
        (
            (
                "There is nothing wrong with your television set. Do not attempt to"
                " adjust the picture. We are controlling transmission. If we wish to"
                " make it louder, we will bring up the volume. If we wish to make it"
                " softer, we will tune it to a whisper. We will control the horizontal."
                " We will control the vertical. We can roll the image, make it flutter."
                " We can change the focus to a soft blur or sharpen it to crystal"
                " clarity. For the next hour, sit quietly, and we will control all that"
                " you see and hear. We repeat: There is nothing wrong with your"
                " television set."
            ),
            (
                "There is nothing wrong with your television set. Do not attempt to"
                " adjust the\npicture. We are controlling transmission. If we wish to"
                " make it louder, we\nwill bring up the volume. If we wish to make it"
                " softer, we will tune it to a\nwhisper. We will control the"
                " horizontal. We will control the vertical. We can\nroll the image,"
                " make it flutter. We can change the focus to a soft blur or\nsharpen"
                " it to crystal clarity. For the next hour, sit quietly, and we"
                " will\ncontrol all that you see and hear. We repeat: There is nothing"
                " wrong with your\ntelevision set."
            ),
        ),
    ],
)
def test_textwrap(console, in_text, out_text):
    """Text is formatted as expected."""
    assert console.textwrap(in_text) == out_text


@pytest.mark.parametrize(
    ("width", "in_text", "out_text"),
    [
        (20, "This is 27 characters long.", "This is 27\ncharacters long."),
        (
            50,
            "This is 57 characters long. This is 57 characters long.",
            "This is 57 characters long. This is 57 characters\nlong.",
        ),
        (
            80,
            (
                "This is 83 characters long. This is 83 characters long. This is 83"
                " characters long."
            ),
            (
                "This is 83 characters long. This is 83 characters long. This is 83"
                " characters\nlong."
            ),
        ),
        (
            120,
            (
                "This is 144 characters long. This is 144 characters long. This is 144"
                " characters long. This is 144 characters long. This is 144 characters"
                " long."
            ),
            (
                "This is 144 characters long. This is 144 characters long. This is 144"
                " characters long. This is 144 characters long. This\nis 144 characters"
                " long."
            ),
        ),
    ],
)
def test_textwrap_width_override(console, width, in_text, out_text):
    """Width override is respected."""
    assert console.textwrap(in_text, width=width) == out_text



import pytest

from briefcase.console import InputDisabled


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("Value", "Value"),
        ("", "Default"),
    ],
)
def test_text_question(console, value, expected):
    console.values = [value]

    actual = console.input_text(
        prompt="> ",
        default="Default",
    )

    assert actual == expected
    assert console.prompts == ["> "]


def test_disabled(disabled_console):
    """If input is disabled, the default is returned."""
    actual = disabled_console.input_text(
        prompt="> ",
        default="Default",
    )

    assert actual == "Default"
    assert disabled_console.prompts == []


def test_disabled_no_default(disabled_console):
    """If input is disabled and there is no default, an error is raised."""
    with pytest.raises(InputDisabled):
        disabled_console.input_text(
            prompt="> ",
            default=None,
        )

    assert disabled_console.prompts == []



from unittest.mock import MagicMock

import pytest

from briefcase.integrations.virtual_environment import VenvEnvironment


@pytest.fixture
def cwd_tmp_path(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)


def test_recreate_true(mock_tools, tmp_path, cwd_tmp_path):
    """Context manager recreates environment when recreate=True."""
    venv_path = tmp_path / "test_venv"
    env = VenvEnvironment(
        mock_tools,
        path=venv_path,
        recreate=True,
    )

    env.venv_context = MagicMock()

    with env as context:
        assert context == env.venv_context
        env.venv_context.recreate.assert_called_once()


def test_recreate_true_exists(mock_tools, tmp_path, cwd_tmp_path):
    """Context manager recreates environment when recreate=True, even when venv
    exists."""
    venv_path = tmp_path / "test_venv"
    env = VenvEnvironment(
        mock_tools,
        path=venv_path,
        recreate=True,
    )

    env.venv_context = MagicMock()
    env.venv_context.exists.return_value = True

    with env as context:
        assert context == env.venv_context
        env.venv_context.exists.assert_not_called()
        env.venv_context.recreate.assert_called_once()
        env.venv_context.create.assert_not_called()


def test_venv_nonexistent(mock_tools, tmp_path, cwd_tmp_path):
    """Context manager creates environment when it doesn't exist."""
    venv_path = tmp_path / "test_venv"
    env = VenvEnvironment(
        mock_tools,
        path=venv_path,
        recreate=False,
    )

    env.venv_context = MagicMock()
    env.venv_context.exists.return_value = False

    with env as context:
        assert context == env.venv_context
        env.venv_context.exists.assert_called_once()
        env.venv_context.create.assert_called_once()


def test_venv_exists(mock_tools, tmp_path, cwd_tmp_path):
    """Context manager does nothing when environment exists and recreate=False."""
    venv_path = tmp_path / "test_venv"
    env = VenvEnvironment(
        mock_tools,
        path=venv_path,
        recreate=False,
    )

    env.venv_context = MagicMock()
    env.venv_context.exists.return_value = True

    with env as context:
        assert context == env.venv_context
        env.venv_context.exists.assert_called_once()
        env.venv_context.create.assert_not_called()
        env.venv_context.recreate.assert_not_called()


def test_exception_handling(mock_tools, tmp_path, cwd_tmp_path):
    """Context manager handles exceptions properly."""
    venv_path = tmp_path / "test_venv"
    env = VenvEnvironment(
        mock_tools,
        path=venv_path,
        recreate=False,
    )

    env.venv_context = MagicMock()
    env.venv_context.exists.return_value = True

    try:
        with env as context:
            assert context == env.venv_context
            raise ValueError("Test exception")
    except ValueError:
        pass

    env.venv_context.exists.assert_called_once()



from unittest import mock

import pytest

from briefcase.commands.base import BaseCommand
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.macOS.app import macOSAppMixin, macOSCreateMixin
from tests.utils import create_file, create_plist_file


class DummyInstallCommand(macOSAppMixin, macOSCreateMixin, BaseCommand):
    """A dummy command to expose package installation capabilities."""

    command = "install"

    def __init__(self, base_path, **kwargs):
        super().__init__(base_path=base_path / "base_path", **kwargs)


@pytest.fixture
def dummy_command(dummy_console, tmp_path):
    cmd = DummyInstallCommand(
        console=dummy_console,
        base_path=tmp_path,
    )

    cmd.tools.subprocess = mock.MagicMock(spec_set=Subprocess)

    return cmd


@pytest.fixture
def first_app_templated(first_app_config, tmp_path):
    app_path = tmp_path / "base_path/build/first-app/macos/app/First App.app"

    # Create the briefcase.toml file
    create_file(
        tmp_path / "base_path/build/first-app/macos/app/briefcase.toml",
        """
[briefcase]
target_version = "0.3.20"

[paths]
app_packages_path="First App.app/Contents/Resources/app_packages"
support_path="First App.app/Contents/Frameworks"
runtime_path="Python.xcframework/macos-arm64_x86_64/Python.framework"
info_plist_path="First App.app/Contents/Info.plist"
entitlements_path="Entitlements.plist"
""",
    )

    # Create the plist file for the app
    create_plist_file(
        app_path / "Contents/Info.plist",
        {
            "MainModule": "first_app",
        },
    )

    # Create the entitlements file for the app
    create_plist_file(
        tmp_path / "base_path/build/first-app/macos/app/Entitlements.plist",
        {
            "com.apple.security.cs.allow-unsigned-executable-memory": True,
            "com.apple.security.cs.disable-library-validation": True,
        },
    )

    # Create some folders that need to exist.
    (app_path / "Contents/Resources/app_packages").mkdir(parents=True)
    (app_path / "Contents/Frameworks").mkdir(parents=True)

    # Create an installer Distribution.xml
    create_file(
        tmp_path / "base_path/build/first-app/macos/app/installer/Distribution.xml",
        """<?xml?>\n<installer-script></installer-script>""",
    )

    # Create the XCframework Info.plist file, with a deliberately weird min macOS version
    create_plist_file(
        (
            tmp_path
            / "base_path/build/first-app/macos/app/support/Python.xcframework"
            / "macos-arm64_x86_64/Python.framework/Resources/Info.plist"
        ),
        {
            "CFBundleVersion": "3.10.15",
            "MinimumOSVersion": "10.12",
        },
    )

    # Select dmg packaging by default
    first_app_config.packaging_format = "dmg"

    return first_app_config



import pytest

from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseConfigError

from .conftest import DummyCommand


@pytest.fixture
def first_app():
    return AppConfig(
        app_name="first",
        bundle="com.example",
        version="0.0.1",
        description="The first simple app",
        sources=["src/first"],
        license={"file": "LICENSE"},
    )


@pytest.fixture
def second_app():
    return AppConfig(
        app_name="second",
        bundle="com.example",
        version="0.0.2",
        description="The second simple app",
        sources=["src/second"],
        license={"file": "LICENSE"},
    )


@pytest.fixture
def base_command(dummy_console, tmp_path, first_app, second_app):
    return DummyCommand(
        console=dummy_console,
        base_path=tmp_path,
        apps={
            "first": first_app,
            "second": second_app,
        },
    )


def test_finalize_all(base_command, first_app, second_app):
    "A call to finalize verifies host, tools, and finalized all app configs"
    base_command.finalize(apps=base_command.apps.values())

    # The right sequence of things will be done
    assert base_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # App config has been finalized
        ("finalize-app-config", "second"),
    ]

    # Apps are no longer in draft mode
    assert not hasattr(first_app, "__draft__")
    assert not hasattr(second_app, "__draft__")


def test_finalize_single(base_command, first_app, second_app):
    "A call to finalize verifies host, tools, and finalizes a single app config"
    base_command.finalize(apps=[first_app])

    # The right sequence of things will be done
    assert base_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
    ]

    # First app is no longer in draft mode; second is
    assert not hasattr(first_app, "__draft__")
    assert hasattr(second_app, "__draft__")


def test_finalize_all_repeat(base_command, first_app, second_app):
    """Multiple calls to finalize verifies host & tools multiple times, but only once on
    "config."""
    # Finalize apps twice. This is an approximation of what happens
    # when a command chain is executed; create, update, build and run will
    # all finalize; create will finalize the app configs, each command will
    # have it's own tools verified.
    base_command.finalize(apps=base_command.apps.values())
    base_command.finalize(apps=base_command.apps.values())

    # The right sequence of things will be done
    assert base_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # App config has been finalized
        ("finalize-app-config", "second"),
        # Host OS is verified again
        ("verify-host",),
        # Tools are verified again
        ("verify-tools",),
    ]

    # Apps are no longer in draft mode
    assert not hasattr(first_app, "__draft__")
    assert not hasattr(second_app, "__draft__")


def test_finalize_single_repeat(base_command, first_app, second_app):
    """Multiple calls to finalize verifies host & tools multiple times, but finalizes
    app "config once."""

    # Finalize app twice. This is an approximation of what happens
    # when a command chain is executed; create, update, build and run will
    # all finalize; create will finalize the app config, each command will
    # have it's own tools verified.
    base_command.finalize(apps=[first_app])
    base_command.finalize(apps=[first_app])

    # The right sequence of things will be done
    assert base_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # Host OS is verified again
        ("verify-host",),
        # Tools are verified again
        ("verify-tools",),
    ]

    # First app is no longer in draft mode; second is
    assert not hasattr(first_app, "__draft__")
    assert hasattr(second_app, "__draft__")


def test_external_and_internal(base_command, first_app):
    """If an app provides both sources and external_package_path, an error is raised."""
    first_app.external_package_path = "path/to/package"

    with pytest.raises(
        BriefcaseConfigError,
        match=r"'first' is declared as an external app, but also defines 'sources'",
    ):
        base_command.finalize(apps=[first_app])


def test_not_external_or_internal(base_command, first_app):
    """If an app provides neither sources or external_package_path, an error is
    raised."""
    first_app.sources = None

    with pytest.raises(
        BriefcaseConfigError,
        match=r"'first' does not define either 'sources' or 'external_package_path'.",
    ):
        base_command.finalize(apps=[first_app])


def test_binary_path_internal_app(base_command, first_app):
    """If an internal app defines external_package_executable_path, an error is
    raised."""
    first_app.external_package_executable_path = "internal/app.exe"

    with pytest.raises(
        BriefcaseConfigError,
        match=(
            r"'first' defines 'external_package_executable_path', "
            r"but not 'external_package_path'"
        ),
    ):
        base_command.finalize(apps=[first_app])



import subprocess
import sys
from unittest import mock

import pytest

from briefcase.integrations.subprocess import Subprocess
from briefcase.integrations.virtual_environment import VenvContext
from briefcase.platforms.windows.app import WindowsAppDevCommand


@pytest.fixture
def dev_command(dummy_console, tmp_path):
    """Fixture for a WindowsAppDevCommand instance."""
    command = WindowsAppDevCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.home_path = tmp_path / "home"
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command._stream_app_logs = mock.MagicMock()
    return command


@pytest.fixture
def default_env() -> VenvContext:
    """Create a venv mock for tests that require a venv parameter."""
    mock_venv = mock.MagicMock(spec=VenvContext)
    mock_venv.run.return_value = mock.MagicMock()
    mock_venv.check_output.return_value = ""
    mock_venv.Popen.return_value = mock.MagicMock()
    mock_venv.executable = "/mock/venv/Scripts/python.exe"
    return mock_venv


def test_dev_app_starts(dev_command, first_app_config, tmp_path, default_env):
    """A Windows app can be started in development mode using Python."""
    log_popen = mock.MagicMock()
    default_env.Popen.return_value = log_popen

    # Run the app
    dev_command.run_dev_app(first_app_config, env={}, venv=default_env, passthrough=[])

    # Extract the actual Popen call arguments
    popen_args, popen_kwargs = default_env.Popen.call_args

    # Check that the command uses the Python executable
    assert popen_args[0][0] == sys.executable

    # The inline Python command should reference the app's main module
    assert "runpy.run_module" in popen_args[0][2]
    assert first_app_config.module_name in popen_args[0][2]

    # Verify other subprocess parameters
    assert popen_kwargs["cwd"] == tmp_path / "home"
    assert popen_kwargs["encoding"] == "UTF-8"
    assert popen_kwargs["stdout"] == subprocess.PIPE
    assert popen_kwargs["stderr"] == subprocess.STDOUT
    assert popen_kwargs["bufsize"] == 1

    # The log streamer should be called
    dev_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        clean_output=False,
    )



from typing import TypedDict


class AppPathMappings(TypedDict):
    device_sys_path_regex: str
    device_subfolders: list[str]
    host_folders: list[str]


class AppPackagesPathMappings(TypedDict):
    sys_path_regex: str
    host_folder: str


class DebuggerConfig(TypedDict):
    debugger: str
    host: str
    port: int
    host_os: str
    app_path_mappings: AppPathMappings | None
    app_packages_path_mappings: AppPackagesPathMappings | None



import shutil
from unittest import mock

import pytest

from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.macOS.app import macOSAppPackageCommand


@pytest.fixture
def package_command(dummy_console, tmp_path):
    command = macOSAppPackageCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )

    command._command_factory = mock.MagicMock()
    command.select_identity = mock.MagicMock()
    command.sign_app = mock.MagicMock()
    command.sign_file = mock.MagicMock()
    command.notarize = mock.MagicMock()
    command.dmgbuild = mock.MagicMock()
    command.tools.subprocess = mock.MagicMock(spec=Subprocess)
    command.verify_not_on_icloud = mock.MagicMock()

    return command


@pytest.fixture
def external_first_app(first_app_with_binaries, first_app_config, tmp_path):
    # Convert the first_app_config into an external app
    first_app_config.sources = None
    first_app_config.external_package_path = (
        tmp_path / "base_path/external/First App.app"
    )

    # Move the binaries from the compiled first app to the external location
    first_app_config.external_package_path.parent.mkdir(parents=True)
    shutil.move(
        tmp_path / "base_path/build/first-app/macos/app/First App.app",
        tmp_path / "base_path/external/First App.app",
    )

    # And purge the original package path, so we have a "clean" template directory
    shutil.rmtree(tmp_path / "base_path/build/first-app/macos/app")

    return first_app_config



import pytest

from briefcase.exceptions import InputDisabled


def test_unvalidated_input(console):
    """If the user enters text and there's no validation, the text is returned."""
    console.values = ["hello"]

    value = console.text_question(
        intro="Some introduction",
        description="My variable",
        default="goodbye",
    )

    assert console.prompts == ["My variable [goodbye]: "]
    assert value == "hello"


def test_unvalidated_input_with_override(console):
    """If an override is provided and there's no validation, the override is
    returned."""
    console.values = ["hello"]

    value = console.text_question(
        intro="Some introduction",
        description="My variable",
        default="goodbye",
        override_value="override",
    )

    assert console.prompts == []
    assert value == "override"


def test_validated_input(console):
    """If the user enters text and there's validation, the user is prompted until valid
    text is entered."""
    console.values = ["bad", "hello"]

    def validator(text):
        if text == "bad":
            raise ValueError("That's bad...")
        return True

    value = console.text_question(
        intro="Some introduction",
        description="My variable",
        default="goodbye",
        validator=validator,
    )

    assert console.prompts == [
        "My variable [goodbye]: ",
        "My variable [goodbye]: ",
    ]
    assert value == "hello"


def test_validated_input_with_override(console):
    """If an override is provided and there's validation, the override is validated and
    returned."""

    def validator(text):
        if text == "bad":
            raise ValueError("That's bad...")
        return True

    value = console.text_question(
        intro="Some introduction",
        description="My variable",
        default="goodbye",
        validator=validator,
        override_value="override",
    )

    assert console.prompts == []
    assert value == "override"


def test_validated_input_with_invalid_override(console):
    """If an invalid override is provided and there's validation, the override is
    rejected and the user is prompted until valid text is entered."""
    console.values = ["bad", "hello"]

    value = console.text_question(
        intro="Some introduction",
        description="My variable",
        default="goodbye",
        validator=lambda text: text not in {"bad", "bad-override"},
        override_value="bad-override",
    )

    assert console.prompts == [
        "My variable [goodbye]: ",
        "My variable [goodbye]: ",
    ]
    assert value == "hello"


def test_input_with_default(console):
    """If the user enters text and there's no validation, the text is returned."""
    console.values = [""]

    value = console.text_question(
        intro="Some introduction", description="My variable", default="goodbye"
    )

    assert console.prompts == ["My variable [goodbye]: "]
    assert value == "goodbye"


def test_input_disabled(disabled_console):
    """If input is disabled, the default is returned."""

    value = disabled_console.text_question(
        intro="Some introduction",
        description="My variable",
        default="goodbye",
    )

    assert disabled_console.prompts == []
    assert value == "goodbye"


def test_input_disabled_with_override(disabled_console):
    """If input is disabled, the override is returned."""
    value = disabled_console.text_question(
        intro="Some introduction",
        description="My variable",
        default="goodbye",
        override_value="override",
    )

    assert disabled_console.prompts == []
    assert value == "override"


def test_input_disabled_validation_failure(disabled_console):
    """If input is disabled, and validation fails, an error is raised."""

    def not_valid(text):
        raise ValueError("Well that won't work...")

    with pytest.raises(InputDisabled, match=r"Well that won't work..."):
        disabled_console.text_question(
            intro="Some introduction",
            description="My variable",
            default="goodbye",
            validator=not_valid,
        )

    assert disabled_console.prompts == []


def test_input_disabled_validation_failure_with_override(disabled_console):
    """If input is disabled, and validation fails for override, an error is raised."""

    def not_valid(text):
        raise ValueError("Well that won't work...")

    with pytest.raises(InputDisabled, match=r"Well that won't work..."):
        disabled_console.text_question(
            intro="Some introduction",
            description="My variable",
            default="goodbye",
            validator=not_valid,
            override_value="override",
        )

    assert disabled_console.prompts == []



import re
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import (
    BriefcaseCommandError,
)
from briefcase.integrations.java import JDK


@pytest.mark.parametrize(
    ("host_os", "java_home"),
    [
        ("Linux", Path("tools", "java17")),
        ("Windows", Path("tools", "java17")),
        ("Darwin", Path("tools", "java17", "Contents", "Home")),
    ],
)
def test_java_running(mock_tools, host_os, java_home, tmp_path):
    """If Java is running, uninstall will raise an error message."""
    # Mock os
    mock_tools.host_os = host_os

    # Create a mock of a previously installed Java version.
    (tmp_path / java_home / "bin").mkdir(parents=True)

    # Create an JDK wrapper
    jdk = JDK(mock_tools, java_home=tmp_path / java_home)

    # Mock shutil.rmtree so that it gives a permission error
    mock_tools.shutil.rmtree.side_effect = PermissionError(
        "File is being used by other processes"
    )

    # Set correct path according to each OS
    if host_os == "Darwin":
        expected_loc = (tmp_path / java_home).parent.parent
    else:
        expected_loc = tmp_path / java_home

    # Attempting to uninstall JDK which will fail due to Permission Error
    with pytest.raises(
        BriefcaseCommandError,
        match=re.escape(f"Permission denied when trying to remove {expected_loc!r}"),
    ):
        jdk.uninstall()


@pytest.mark.parametrize(
    ("host_os", "java_home"),
    [
        ("Linux", Path("tools", "java17")),
        ("Windows", Path("tools", "java17")),
        ("Darwin", Path("tools", "java17", "Contents", "Home")),
    ],
)
def test_java_not_running(mock_tools, host_os, java_home, tmp_path):
    """If Java is not running, uninstalling should proceed smoothly."""
    # Mock os
    mock_tools.host_os = host_os

    # Create a mock of a previously installed Java version.
    (tmp_path / java_home / "bin").mkdir(parents=True)

    # Create an JDK wrapper
    jdk = JDK(mock_tools, java_home=tmp_path / java_home)

    # Set correct path according to each OS
    if host_os == "Darwin":
        expected_loc = (tmp_path / java_home).parent.parent
    else:
        expected_loc = tmp_path / java_home

    # Mock so that shutil behave normally
    mock_tools.shutil.rmtree = MagicMock()

    # Uninstall JDK
    jdk.uninstall()

    # Verify that shutil works normally and that correct location is passed
    mock_tools.shutil.rmtree.assert_called_once_with(expected_loc)



import sys

import pytest

from briefcase.platforms.macOS import utils


@pytest.mark.skipif(sys.platform != "darwin", reason="Test runs only on macOS")
@pytest.mark.parametrize(
    ("mime_type", "uti"),
    [
        (None, None),
        ("application/pdf", "com.adobe.pdf"),
        ("text/plain", "public.plain-text"),
        ("image/png", "public.png"),
        ("application/unknown", None),
    ],
)
def test_mime_type_to_uti(mime_type, uti):
    """Check if a MIME type can be converted to a UTI."""
    assert utils.mime_type_to_uti(mime_type) == uti


@pytest.mark.skipif(sys.platform != "darwin", reason="Test runs only on macOS")
def test_mime_type_to_uti_with_nonexisting_coretypes_file(monkeypatch):
    """Test that mime_type_to_UTI returns None if the coretypes file doesn't exist."""
    monkeypatch.setattr(utils, "CORETYPES_PATH", "/does/not/exist")
    assert utils.mime_type_to_uti("application/pdf") is None



from importlib.metadata import entry_points

from briefcase.debuggers.base import BaseDebugger
from briefcase.debuggers.debugpy import DebugpyDebugger  # noqa: F401
from briefcase.debuggers.pdb import PdbDebugger  # noqa: F401
from briefcase.exceptions import BriefcaseCommandError


def get_debuggers() -> dict[str, type[BaseDebugger]]:
    """Loads built-in and third-party debuggers."""
    return {
        entry_point.name: entry_point.load()
        for entry_point in entry_points(group="briefcase.debuggers")
    }


def get_debugger(name: str) -> BaseDebugger:
    """Get a debugger by name."""
    debuggers = get_debuggers()
    if name not in debuggers:
        raise BriefcaseCommandError(f"Unknown debugger: {name}")
    return debuggers[name]()



import pytest

from briefcase.exceptions import MissingToolError
from briefcase.integrations.wix import WiX


def test_non_existing_wix_install(mock_tools, tmp_path):
    """If there's no existing managed WiX install, upgrading is an error."""
    # Create an SDK wrapper around a non-existing managed install
    wix = WiX(mock_tools)

    with pytest.raises(MissingToolError):
        wix.upgrade()

    # No download was attempted
    assert mock_tools.file.download.call_count == 0


def test_wix_uninstall(mock_tools, tmp_path):
    """The uninstall method removes a managed install."""
    # Create a mock of a previously installed WiX version.
    wix_path = tmp_path / "tools/wix"
    wix_path.mkdir(parents=True)

    wix = WiX(mock_tools)
    wix.uninstall()
    mock_tools.shutil.rmtree.assert_called_once_with(wix_path)



import pytest

from ...utils import create_file


# Windows' AppConfig requires attribute 'packaging_format'
@pytest.fixture
def first_app_config(first_app_config):
    first_app_config.packaging_format = "msi"
    return first_app_config


@pytest.fixture
def external_first_app(first_app_config, tmp_path):
    first_app_config = first_app_config.copy()
    first_app_config.sources = None
    first_app_config.external_package_path = tmp_path / "base_path/external/src"
    first_app_config.external_package_executable_path = "internal/app.exe"

    # Create the binary
    create_file(
        tmp_path / "base_path/external/src/internal/app.exe",
        "external binary",
    )

    return first_app_config



import subprocess
from unittest.mock import MagicMock, call

import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.platforms.linux.system import LinuxSystemBuildCommand


@pytest.fixture
def build_command(dummy_console, tmp_path, first_app):
    command = LinuxSystemBuildCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
        apps={"first": first_app},
    )
    command.tools.host_os = "Linux"
    command.tools.host_arch = "wonky"

    # All calls to `shutil.which()` succeed
    command.tools.shutil.which = MagicMock(return_value="/path/to/exe")

    # Mock subprocess
    command.tools.subprocess = MagicMock()

    return command


def test_deb_requirements(build_command, first_app_config):
    """Debian requirements can be verified."""
    first_app_config.target_vendor_base = "debian"

    build_command.verify_system_packages(first_app_config)

    # The packages were verified
    assert build_command.tools.subprocess.check_output.mock_calls == [
        call(["dpkg", "-s", "python3-dev"], quiet=1),
        call(["dpkg", "-s", "dpkg-dev"], quiet=1),
        call(["dpkg", "-s", "g++"], quiet=1),
        call(["dpkg", "-s", "gcc"], quiet=1),
        call(["dpkg", "-s", "libc6-dev"], quiet=1),
        call(["dpkg", "-s", "make"], quiet=1),
    ]


def test_rpm_requirements(build_command, first_app_config):
    """RHEL requirements can be verified."""
    first_app_config.target_vendor_base = "rhel"

    build_command.verify_system_packages(first_app_config)

    assert build_command.tools.subprocess.check_output.mock_calls == [
        call(["rpm", "-q", "python3-devel"], quiet=1),
        call(["rpm", "-q", "gcc"], quiet=1),
        call(["rpm", "-q", "make"], quiet=1),
        call(["rpm", "-q", "pkgconf-pkg-config"], quiet=1),
    ]


def test_suse_requirements(build_command, first_app_config):
    """SUSE requirements can be verified."""
    first_app_config.target_vendor_base = "suse"

    build_command.verify_system_packages(first_app_config)

    assert build_command.tools.subprocess.check_output.mock_calls == [
        call(["rpm", "-q", "--whatprovides", "python3-devel"], quiet=1),
        call(
            ["rpm", "-q", "--whatprovides", "patterns-devel-base-devel_basis"], quiet=1
        ),
    ]


def test_arch_requirements(build_command, first_app_config, capsys):
    """Arch requirements can be verified."""
    first_app_config.target_vendor_base = "arch"

    build_command.verify_system_packages(first_app_config)

    assert build_command.tools.subprocess.check_output.mock_calls == [
        call(["pacman", "-Q", "python3"], quiet=1),
        call(["pacman", "-Q", "base-devel"], quiet=1),
    ]


def test_unknown_requirements(build_command, first_app_config, capsys):
    """An unknown system can't be verified."""
    first_app_config.target_vendor_base = "somevendor"

    build_command.verify_system_packages(first_app_config)

    # No packages verified
    build_command.tools.subprocess.check_output.assert_not_called()

    # A warning was logged.
    output = capsys.readouterr().out
    assert "WARNING: Can't verify system packages" in output


def test_missing_packages(build_command, first_app_config, capsys):
    """If there are missing system packages, an error is raised."""
    # Mock the system requirement tools; there's a base requirement of packages called
    # "compiler" and "compiler++", plus 3 packages provided by an installation alias
    # "alias". These packages are verified using "check <pkg>", and installed using
    # "system install_flag <pkg>"
    build_command._system_requirement_tools = MagicMock(
        return_value=(
            [
                "compiler",
                "compiler++",
                # Three packages provided by the `alias` alias
                ("aliased-1", "alias"),
                ("aliased-2", "alias"),
                ("aliased-3", "alias"),
            ],
            None,
            ["check"],
            ["system", "install_flag"],
        )
    )

    # Add some system requirements.
    first_app_config.system_requires = ["first", "second", "third"]

    # Mock the side effect of checking those requirements.
    build_command.tools.subprocess.check_output.side_effect = [
        subprocess.CalledProcessError(cmd="check", returncode=1),  # compiler
        "installed",  # compiler++
        subprocess.CalledProcessError(cmd="check", returncode=1),  # aliased-1
        "installed",  # aliased-2
        subprocess.CalledProcessError(cmd="check", returncode=1),  # aliased-3
        "installed",  # first
        subprocess.CalledProcessError(cmd="check", returncode=1),  # second
        "installed",  # third
    ]

    # Verify the requirements. This will raise an error, but the error
    # message will tell you how to install the system packages. This includes
    # using an alias for the install of aliased-1 and aliased-3.
    with pytest.raises(
        BriefcaseCommandError,
        match=r"    sudo system install_flag alias compiler second",
    ):
        build_command.verify_system_packages(first_app_config)

    # All requirements are checked; base package names are checked for aliases
    assert build_command.tools.subprocess.check_output.mock_calls == [
        call(["check", "compiler"], quiet=1),
        call(["check", "compiler++"], quiet=1),
        call(["check", "aliased-1"], quiet=1),
        call(["check", "aliased-2"], quiet=1),
        call(["check", "aliased-3"], quiet=1),
        call(["check", "first"], quiet=1),
        call(["check", "second"], quiet=1),
        call(["check", "third"], quiet=1),
    ]


def test_missing_system_verify(build_command, first_app_config, capsys):
    """If the program to verify system packages doesn't exist, a warning is logged."""
    # Mock the system verifier is missing
    build_command.tools.shutil.which = MagicMock(return_value="")

    build_command.verify_system_packages(first_app_config)

    # No packages verified
    build_command.tools.subprocess.check_output.assert_not_called()

    # A warning was logged.
    output = capsys.readouterr().out
    assert "WARNING: Can't verify system packages" in output


def test_packages_installed(build_command, first_app_config, capsys):
    """If all required packages are installed, no error is raised."""
    # Mock the system requirement tools; there's a base requirement of
    # a packaged called "compiler", verified using "check <pkg>", and
    # installed using "system <pkg>"
    build_command._system_requirement_tools = MagicMock(
        return_value=(
            ["compiler"],
            None,
            ["check"],
            ["system", "install_flag"],
        )
    )

    # Add some system requirements.
    first_app_config.system_requires = ["first", "second", "third"]
    first_app_config.system_runtime_requires = ["first", "other", "compiler"]

    # Mock the effect of checking requirements that are all present
    build_command.tools.subprocess.check_output.return_value = "installed"

    # Verify the requirements. This will raise an error.
    build_command.verify_system_packages(first_app_config)

    # All requirements are checked, but `first` and `compiler` are only verified once
    assert build_command.tools.subprocess.check_output.mock_calls == [
        call(["check", "compiler"], quiet=1),
        call(["check", "first"], quiet=1),
        call(["check", "second"], quiet=1),
        call(["check", "third"], quiet=1),
        call(["check", "other"], quiet=1),
    ]


def test_missing_virtual_packages(build_command, first_app_config, capsys):
    """If there are missing virtual system packages, an error is raised."""
    # Mock the system requirement tools; there's a base requirement of packages called
    # "compiler" and "compiler++", plus 3 packages provided by an installation alias
    # "alias". These packages are verified using "check <pkg>", and installed using
    # "system install_flag <pkg>"

    def devirtualize(package):
        if package.endswith("-virtual"):
            return package[:-8]
        return None

    build_command._system_requirement_tools = MagicMock(
        return_value=(
            ["compiler"],
            devirtualize,
            ["check"],
            ["system", "install_flag"],
        )
    )

    # Add some system requirements. `second` and `third` are virtual requirements
    first_app_config.system_requires = ["first", "second-virtual", "third-virtual"]

    # Mock the side effect of checking those requirements. third is available in it's
    # devirtualized form. first isn't available, but isn't a virtual package.
    build_command.tools.subprocess.check_output.side_effect = [
        "installed",  # compiler
        subprocess.CalledProcessError(cmd="check", returncode=1),  # first
        subprocess.CalledProcessError(cmd="check", returncode=1),  # second-virtual
        subprocess.CalledProcessError(cmd="check", returncode=1),  # second
        subprocess.CalledProcessError(cmd="check", returncode=1),  # third-virtual
        "installed",  # third
    ]

    # Verify the requirements. This will raise an error, but the error message will tell
    # you how to install the system packages. This includes using an alias for the
    # install of aliased-1, and the original non-virtualized name for `second-virtual`
    with pytest.raises(
        BriefcaseCommandError,
        match=r"    sudo system install_flag first second-virtual",
    ):
        build_command.verify_system_packages(first_app_config)

    # All requirements are checked; base package names are checked for
    # both aliases and virtual packages
    assert build_command.tools.subprocess.check_output.mock_calls == [
        call(["check", "compiler"], quiet=1),
        call(["check", "first"], quiet=1),
        # Second is checked in both virtual and devirtualized form
        call(["check", "second-virtual"], quiet=1),
        call(["check", "second"], quiet=1),
        # Third is checked in both virtual and devirtualized form
        call(["check", "third-virtual"], quiet=1),
        call(["check", "third"], quiet=1),
    ]


def test_virtual_packages_installed(build_command, first_app_config, capsys):
    """If all required packages are installed, no error is raised."""
    # Mock the system requirement tools; there's a base requirement of
    # a packaged called "compiler", verified using "check <pkg>", and
    # installed using "system <pkg>", and `devirtualize` is used as
    # to devirtualize package names.

    def devirtualize(package):
        if package.endswith("-virtual"):
            return package[:-8]
        return None

    build_command._system_requirement_tools = MagicMock(
        return_value=(
            ["compiler"],
            devirtualize,
            ["check"],
            ["system", "install_flag"],
        )
    )

    # Add some system requirements; two of them are virtual packages
    first_app_config.system_requires = ["first", "second", "third-virtual"]
    first_app_config.system_runtime_requires = ["first", "other-virtual"]

    # Mock the effect of checking requirements that are all present,
    # although the virtual packages are only found in devirtualized form.
    build_command.tools.subprocess.check_output.side_effect = [
        "installed",  # compiler
        "installed",  # first
        "installed",  # second
        subprocess.CalledProcessError(cmd="check", returncode=1),  # third-virtual
        "installed",  # third-virtual
        subprocess.CalledProcessError(cmd="check", returncode=1),  # other-virtual
        "installed",  # other
    ]

    # Verify the requirements. This will raise an error.
    build_command.verify_system_packages(first_app_config)

    # All requirements are checked, but `first` is only verified once
    assert build_command.tools.subprocess.check_output.mock_calls == [
        call(["check", "compiler"], quiet=1),
        call(["check", "first"], quiet=1),
        call(["check", "second"], quiet=1),
        # Third is checked in both virtual and devirtualized form
        call(["check", "third-virtual"], quiet=1),
        call(["check", "third"], quiet=1),
        # Other is checked in both virtual and devirtualized form
        call(["check", "other-virtual"], quiet=1),
        call(["check", "other"], quiet=1),
    ]
